# coding: utf-8

import tkinter as tk
from tkinter.messagebox import showerror
from _tkinter import TclError
from platform import system
import numpy as np
from time import time, sleep
from typing import Optional, Tuple
from pkg_resources import resource_string
from io import BytesIO
import logging
from multiprocessing import current_process, Event, Queue
from multiprocessing.queues import Queue as MPQueue

from .config_tools import Zoom, HistogramProcess
from ...camera.meta_camera.camera_setting import CameraBoolSetting, \
  CameraChoiceSetting, CameraScaleSetting
from ...camera.meta_camera import Camera
from ..._global import OptionalModule

try:
  from PIL import ImageTk, Image
except (ModuleNotFoundError, ImportError):
  ImageTk = OptionalModule("pillow")
  Image = OptionalModule("pillow")


class CameraConfig(tk.Tk):
  """This class is a GUI allowing the user to visualize the images from a
  :class:`~crappy.camera.Camera` before a Crappy test starts, and to tune the
  settings of the Camera.

  It is meant to be user-friendly and interactive. It is possible to zoom on
  the image using the mousewheel, and to move on the zoomed image by
  right-clicking and dragging.

  In addition to the image, the interface also displays a histogram of the
  pixel values, an FPS counter, a detected bits counter, the minimum and
  maximum pixel values, and the value and position of the pixel currently under
  the mouse. A checkbox allows auto-adjusting the pixel range to get a better
  contrast.

  This class is used as is by the :class:`~crappy.blocks.Camera`, but also 
  subclassed to provide more specific functionalities to other camera-related 
  Blocks like :class:`~crappy.blocks.VideoExtenso` or 
  :class:`~crappy.blocks.DICVE`.
  
  This class is a child of :obj:`tkinter.Tk`. It relies on the 
  :class:`~crappy.tool.camera_config.config_tools.Zoom` and 
  :class:`~crappy.tool.camera_config.config_tools.HistogramProcess` tools. It
  also interacts with instances of the 
  :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting` class.

  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Camera_config to CameraConfig
  """

  def __init__(self,
               camera: Camera,
               log_queue: MPQueue,
               log_level: Optional[int],
               max_freq: Optional[float]) -> None:
    """Initializes the interface and displays it.

    Args:
      camera: The :class:`~crappy.camera.Camera` object in charge of acquiring 
        the images.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to 
        the main :obj:`~logging.Logger`, only used in Windows.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.
      max_freq: The maximum frequency this window is allowed to loop at. It is
        simply the ``freq`` attribute of the :class:`~crappy.blocks.Camera`
        Block.
    
    .. versionadded:: 2.0.0 *log_queue*, *log_level* and *max_freq* arguments
    """

    super().__init__()
    self._camera = camera
    self.shape = None
    self.dtype = None
    self._logger: Optional[logging.Logger] = None

    # Instantiating objects for the process managing the histogram calculation
    self._stop_event = Event()
    self._processing_event = Event()
    self._img_in = Queue(maxsize=0)
    self._img_out = Queue(maxsize=0)
    self._histogram_process = HistogramProcess(
        stop_event=self._stop_event, processing_event=self._processing_event,
        img_in=self._img_in, img_out=self._img_out, log_level=log_level,
        log_queue=log_queue)

    # Attributes containing the several images and histograms
    self._img = None
    self._pil_img = None
    self._original_img = None
    self._hist = None
    self._pil_hist = None

    # Other attributes used in this class
    self._low_thresh = None
    self._high_thresh = None
    self._move_x = None
    self._move_y = None
    self._run = True
    self._n_loops = 0
    self._max_freq = max_freq

    # Settings for adjusting the behavior of the zoom
    self._zoom_ratio = 0.9
    self._zoom_step = 0
    self._max_zoom_step = 15

    # Settings of the root window
    self.title(f'Configuration window for the camera: {type(camera).__name__}')
    self.protocol("WM_DELETE_WINDOW", self.finish)
    self._zoom_values = Zoom()

    # Initializing the interface
    self._set_variables()
    self._set_traces()
    self._set_layout()
    self._set_bindings()
    self._add_settings()
    self.update()

  def main(self) -> None:
    """Constantly updates the image and the information on the GUI, until asked
    to stop.
    
    .. versionadded:: 1.5.10
    """

    # Starting the histogram calculation process
    self._histogram_process.start()

    self._n_loops = 0
    start_time = time()

    while self._run:
      # Remaining below the max allowed frequency
      if self._max_freq is None or (self._n_loops <
                                    self._max_freq * (time() - start_time)):
        # Update the image, the histogram and the information
        self._update_img()

      # Update the FPS counter
      if time() - start_time > 0.5:
        self._fps_var.set(self._n_loops / (time() - start_time))
        self._n_loops = 0
        start_time = time()

  def log(self, level: int, msg: str) -> None:
    """Record log messages for the CameraConfig window.

    Also instantiates the :obj:`~logging.Logger` when logging the first
    message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    
    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{type(self).__name__}")

    self._logger.log(level, msg)

  def report_callback_exception(self, exc: Exception, val: str, tb) -> None:
    """Method displaying an error message in case an exception is raised in a
    :mod:`tkinter` callback.

    .. versionadded:: 2.0.0
    """

    self._logger.exception(f"Caught exception in {type(self).__name__}: "
                           f"{exc.__name__}({val})", exc_info=tb)
    showerror("Error !", message=f"{exc.__name__}\n{val}")

  def finish(self) -> None:
    """Method called when the user tries to close the configuration window.

    Mostly intended for being overwritten.
    
    .. versionadded:: 2.0.0
    """

    self.stop()

  def stop(self) -> None:
    """Method called for gracefully stopping the GUI.

    Stops the process calculating the histogram, and destroys the GUI.

    .. versionadded:: 2.0.0
    """

    # Stopping the event loop and the histogram process
    self._run = False
    self._stop_event.set()
    sleep(0.1)

    # Killing the histogram process if it's still alive
    if self._histogram_process.is_alive():
      self.log(logging.WARNING, "The histogram process failed to stop, "
                                "killing it !")
      self._histogram_process.terminate()

    self.log(logging.DEBUG, "Destroying the configuration window")

    try:
      self.destroy()
    except TclError:
      self.log(logging.WARNING, "Cannot destroy the configuration window, "
                                "ignoring")

  def _set_layout(self) -> None:
    """Creates and places the different elements of the display on the GUI."""

    self.log(logging.DEBUG, "Setting the interface layout")

    # The main frame of the window
    self._main_frame = tk.Frame()
    self._main_frame.pack(fill='both', expand=True)

    # The frame containing the image and the histogram
    self._graphical_frame = tk.Frame(self._main_frame)
    self._graphical_frame.pack(expand=True, fill="both", anchor="w",
                               side="left")

    # The image row will expand 4 times as fast as the histogram row
    self._graphical_frame.columnconfigure(0, weight=1)
    self._graphical_frame.rowconfigure(0, weight=1)
    self._graphical_frame.rowconfigure(1, weight=4)

    # Adapting the default dimension of the GUI according to the screen size
    screen_width = self.winfo_screenwidth()
    screen_height = self.winfo_screenheight()
    if screen_width < 1600 or screen_height < 900:
      min_width, min_height = 600, 450
    else:
      min_width, min_height = 800, 600

    # The label containing the histogram
    self._hist_canvas = tk.Canvas(self._graphical_frame, height=80,
                                  width=min_width, highlightbackground='black',
                                  highlightthickness=1)
    self._hist_canvas.grid(row=0, column=0, sticky='nsew')

    # The label containing the image
    self._img_canvas = tk.Canvas(self._graphical_frame, width=min_width,
                                 height=min_height)
    self._img_canvas.grid(row=1, column=0, sticky='nsew')

    # The frame containing the information on the image and the settings
    self._text_frame = tk.Frame(self._main_frame, highlightbackground='black',
                                highlightthickness=1)
    self._text_frame.pack(expand=True, fill='y', anchor='ne')

    # The frame containing the information on the image
    self._info_frame = tk.Frame(self._text_frame, highlightbackground='black',
                                highlightthickness=1)
    self._info_frame.pack(expand=False, fill='both', anchor='n', side='top',
                          ipady=2)

    # The information on the image
    self._fps_label = tk.Label(self._info_frame, textvariable=self._fps_txt)
    self._fps_label.pack(expand=False, fill='none', anchor='n', side='top')

    self._auto_range_button = tk.Checkbutton(self._info_frame,
                                             text='Auto range',
                                             variable=self._auto_range)
    self._auto_range_button.pack(expand=False, fill='none', anchor='n',
                                 side='top')

    self._auto_apply_button = tk.Checkbutton(self._info_frame,
                                             text='Auto apply',
                                             variable=self._auto_apply)
    self._auto_apply_button.pack(expand=False, fill='none', anchor='n',
                                 side='top')

    self._min_max_label = tk.Label(self._info_frame,
                                   textvariable=self._min_max_pix_txt)
    self._min_max_label.pack(expand=False, fill='none', anchor='n', side='top')

    self._bits_label = tk.Label(self._info_frame, textvariable=self._bits_txt)
    self._bits_label.pack(expand=False, fill='none', anchor='n', side='top')

    self._zoom_label = tk.Label(self._info_frame, textvariable=self._zoom_txt)
    self._zoom_label.pack(expand=False, fill='none', anchor='n', side='top')

    self._reticle_label = tk.Label(self._info_frame,
                                   textvariable=self._reticle_txt)
    self._reticle_label.pack(expand=False, fill='none', anchor='n', side='top')

    # The frame containing the settings, the message and the update button
    self._sets_frame = tk.Frame(self._text_frame)
    self._sets_frame.pack(expand=True, fill='both', anchor='e', side='top')

    # Tha label warning the user
    self._validate_text = tk.Label(
      self._sets_frame,
      text='To validate the choice of the settings and start the test, simply '
           'close this window.',
      fg='#f00', wraplength=300)
    self._validate_text.pack(expand=False, fill='none', ipadx=5, ipady=5,
                             padx=5, pady=5, anchor='n', side='top')

    # The update button
    self._create_buttons()

    # The frame containing the settings
    self._settings_frame = tk.Frame(self._sets_frame,
                                    highlightbackground='black',
                                    highlightthickness=1)
    self._settings_frame.pack(expand=True, fill='both', anchor='n', side='top')

    # The canvas containing the settings
    self._settings_canvas = tk.Canvas(self._settings_frame)
    self._settings_canvas.pack(expand=True, fill='both', anchor='w',
                               side='left')
    self._canvas_frame = tk.Frame(self._settings_canvas)
    self._id = self._settings_canvas.create_window(
      0, 0, window=self._canvas_frame, anchor='nw',
      width=self._settings_canvas.winfo_reqwidth(), tags='canvas window')

    # Creating the scrollbar
    self._vbar = tk.Scrollbar(self._settings_frame, orient="vertical")
    self._vbar.pack(expand=True, fill='y', side='right')
    self._vbar.config(command=self._custom_yview)

    # Associating the scrollbar with the settings canvas
    self._settings_canvas.config(yscrollcommand=self._vbar.set)

  def _create_buttons(self) -> None:
    """This method is meant to simplify the addition of extra buttons in
    subclasses."""

    self._update_button = tk.Button(self._sets_frame, text="Apply Settings",
                                    command=self._update_settings)
    self._update_button.pack(expand=False, fill='none', ipadx=5, ipady=5,
                             padx=5, pady=5, anchor='n', side='top')

  def _custom_yview(self, *args) -> None:
    """Custom handling of the settings canvas scrollbar, that does nothing
    if the entire canvas is already visible."""

    if self._settings_canvas.yview() == (0., 1.):
      return
    self._settings_canvas.yview(*args)

  def _set_bindings(self) -> None:
    """Sets the bindings for the different events triggered by the user."""

    self.log(logging.DEBUG, "Setting the interface bindings")

    # Bindings for the settings canvas
    self._settings_canvas.bind("<Configure>", self._configure_canvas)
    self._settings_frame.bind('<Enter>', self._bind_mouse)
    self._settings_frame.bind('<Leave>', self._unbind_mouse)

    # Different mousewheel handling depending on the platform
    if system() == "Linux":
      self._img_canvas.bind('<4>', self._on_wheel_img)
      self._img_canvas.bind('<5>', self._on_wheel_img)
    else:
      self._img_canvas.bind('<MouseWheel>', self._on_wheel_img)

    # Bindings for the image canvas
    self._img_canvas.bind('<Motion>', self._update_coord)
    self._img_canvas.bind('<ButtonPress-3>', self._start_move)
    self._img_canvas.bind('<B3-Motion>', self._move)

    # It's more efficient to bind the resizing to the graphical frame
    self._graphical_frame.bind("<Configure>", self._on_img_resize)
    self._graphical_frame.bind("<Configure>", self._on_hist_resize)

  def _bind_mouse(self, _: tk.Event) -> None:
    """Binds the mousewheel to the settings canvas scrollbar when the user
    hovers over the canvas."""

    self.log(logging.DEBUG, "Binding the mouse to the image canvas")

    if system() == "Linux":
      self._settings_frame.bind_all('<4>', self._on_wheel_settings)
      self._settings_frame.bind_all('<5>', self._on_wheel_settings)
    else:
      self._settings_frame.bind_all('<MouseWheel>', self._on_wheel_settings)

  def _unbind_mouse(self, _: tk.Event) -> None:
    """Unbinds the mousewheel to the settings canvas scrollbar when the mouse
    leaves the canvas."""

    self.log(logging.DEBUG, "Unbinding the mouse from the image canvas")

    self._settings_frame.unbind_all('<4>')
    self._settings_frame.unbind_all('<5>')
    self._settings_frame.unbind_all('<MouseWheel>')

  def _configure_canvas(self, event: tk.Event) -> None:
    """Adjusts the size of the scrollbar according to the size of the settings
    canvas whenever it is being resized."""

    self.log(logging.DEBUG, "The image canvas has been resized")

    # Adjusting the height of the settings window inside the canvas
    self._settings_canvas.itemconfig(
      self._id, width=event.width,
      height=self._canvas_frame.winfo_reqheight())

    # Setting the scroll region according to the height of the settings window
    self._settings_canvas.configure(
      scrollregion=(0, 0, self._canvas_frame.winfo_reqwidth(),
                    self._canvas_frame.winfo_reqheight()))

  def _on_wheel_settings(self, event: tk.Event) -> None:
    """Scrolls the canvas up or down upon wheel motion."""

    # Do nothing if the entire canvas is already visible
    if self._settings_canvas.yview() == (0., 1.):
      return

    # Different wheel management in Windows and Linux
    if system() == "Linux":
      delta = 1 if event.num == 4 else -1
    else:
      delta = int(event.delta / abs(event.delta))

    self._settings_canvas.yview_scroll(-delta, "units")

  def _on_wheel_img(self, event: tk.Event) -> None:
    """Zooms in or out on the image upon mousewheel motion.

    Handles the specific cases when the mouse is not on the image, or the
    maximum or minimum zoom levels are reached.
    """

    # If the mouse is on the canvas but not on the image, do nothing
    if not self._check_event_pos(event):
      return

    self.log(logging.DEBUG, "Zooming on the canvas")

    pil_width = self._pil_img.width
    pil_height = self._pil_img.height
    zoom_x_low, zoom_x_high = self._zoom_values.x_low, self._zoom_values.x_high
    zoom_y_low, zoom_y_high = self._zoom_values.y_low, self._zoom_values.y_high

    # Different wheel management in Windows and Linux
    if system() == "Linux":
      delta = 1 if event.num == 4 else -1
    else:
      delta = int(event.delta / abs(event.delta))

    # Handling the cases when the minimum or maximum zoom levels are reached
    self._zoom_step += delta
    if self._zoom_step < 0:
      self._zoom_step = 0
      self._zoom_level.set(100)
      return
    elif self._zoom_step == 0:
      self._zoom_values.reset()
      self._zoom_level.set(100)
      self._on_img_resize()
      return
    elif self._zoom_step > self._max_zoom_step:
      self._zoom_step = self._max_zoom_step
      self._zoom_level.set(100 * (1 / self._zoom_ratio) ** self._max_zoom_step)
      return

    # Correcting the event position to make it relative to the image and not
    # the canvas
    zero_x = (self._img_canvas.winfo_width() - pil_width) / 2
    zero_y = (self._img_canvas.winfo_height() - pil_height) / 2
    corr_x = event.x - zero_x
    corr_y = event.y - zero_y

    # The position of the mouse on the image as a ratio between 0 and 1
    x_ratio = corr_x * (zoom_x_high - zoom_x_low) / pil_width
    y_ratio = corr_y * (zoom_y_high - zoom_y_low) / pil_height

    # Updating the upper and lower limits of the image on the display
    ratio = self._zoom_ratio if delta < 0 else 1 / self._zoom_ratio
    self._zoom_values.update_zoom(x_ratio, y_ratio, ratio)

    # Redrawing the image and updating the information
    self._on_img_resize()
    self._zoom_level.set(100 * (1 / self._zoom_ratio) ** self._zoom_step)

  def _update_coord(self, event: tk.Event) -> None:
    """Updates the coordinates of the pixel pointed by the mouse on the
    image."""

    self.log(logging.DEBUG, "Updating the coordinates of the current pixel")

    # If the mouse is on the canvas but not on the image, do nothing
    if not self._check_event_pos(event):
      return

    x_coord, y_coord = self._coord_to_pix(event.x, event.y)

    self._x_pos.set(x_coord)
    self._y_pos.set(y_coord)

    self._update_pixel_value()

  def _update_pixel_value(self) -> None:
    """Updates the display of the gray level value of the pixel currently being
    pointed by the mouse."""

    self.log(logging.DEBUG, "Updating the value of the current pixel")

    try:
      self._reticle_val.set(np.average(self._original_img[self._y_pos.get(),
                                                          self._x_pos.get()]))
    except IndexError:
      self._x_pos.set(0)
      self._y_pos.set(0)
      self._reticle_val.set(np.average(self._original_img[self._y_pos.get(),
                                                          self._x_pos.get()]))

  def _coord_to_pix(self, x: int, y: int) -> Tuple[int, int]:
    """Converts the coordinates of the mouse in the GUI referential to
    coordinates on the original image."""

    pil_width = self._pil_img.width
    pil_height = self._pil_img.height
    zoom_x_low, zoom_x_high = self._zoom_values.x_low, self._zoom_values.x_high
    zoom_y_low, zoom_y_high = self._zoom_values.y_low, self._zoom_values.y_high
    img_height, img_width, *_ = self._img.shape

    # Correcting the event position to make it relative to the image and not
    # the canvas
    zero_x = (self._img_canvas.winfo_width() - pil_width) / 2
    zero_y = (self._img_canvas.winfo_height() - pil_height) / 2
    corr_x = x - zero_x
    corr_y = y - zero_y

    # Convert the relative coordinate of the mouse on the display to coordinate
    # of the mouse on the original image
    x_disp = corr_x / pil_width * (zoom_x_high - zoom_x_low) * img_width
    y_disp = corr_y / pil_height * (zoom_y_high - zoom_y_low) * img_height

    # The coordinate of the upper left corner of the displayed image
    # (potentially zoomed) on the original image
    x_trim = zoom_x_low * img_width
    y_trim = zoom_y_low * img_height

    return min(int(x_disp + x_trim),
               img_width - 1), min(int(y_disp + y_trim), img_height - 1)

  def _start_move(self, event: tk.Event) -> None:
    """Stores the position of the mouse upon left-clicking on the image."""

    # If the mouse is on the canvas but not on the image, do nothing
    if not self._check_event_pos(event):
      return

    self.log(logging.DEBUG, "Drag started")

    # Stores the position of the mouse relative to the top left corner of the
    # image
    zero_x = (self._img_canvas.winfo_width() - self._pil_img.width) / 2
    zero_y = (self._img_canvas.winfo_height() - self._pil_img.height) / 2
    self._move_x = event.x - zero_x
    self._move_y = event.y - zero_y

  def _move(self, event: tk.Event) -> None:
    """Drags the image upon prolonged left-clik and drag from the user."""

    # If the mouse is on the canvas but not on the image, do nothing
    if not self._check_event_pos(event):
      return

    self.log(logging.DEBUG, "Drag ended")

    pil_width = self._pil_img.width
    pil_height = self._pil_img.height
    zoom_x_low, zoom_x_high = self._zoom_values.x_low, self._zoom_values.x_high
    zoom_y_low, zoom_y_high = self._zoom_values.y_low, self._zoom_values.y_high

    # Getting the position delta, in the coordinates of the display
    zero_x = (self._img_canvas.winfo_width() - pil_width) / 2
    zero_y = (self._img_canvas.winfo_height() - pil_height) / 2
    delta_x_disp = self._move_x - (event.x - zero_x)
    delta_y_disp = self._move_y - (event.y - zero_y)

    # Converting the position delta to a ratio between 0 and 1 relative to the
    # size of the original image
    delta_x = delta_x_disp * (zoom_x_high - zoom_x_low) / pil_width
    delta_y = delta_y_disp * (zoom_y_high - zoom_y_low) / pil_height

    # Actually updating the display
    self._zoom_values.update_move(delta_x, delta_y)

    # Resetting the original position, otherwise the drag never ends
    self._move_x = event.x - zero_x
    self._move_y = event.y - zero_y

  def _check_event_pos(self, event: tk.Event) -> bool:
    """Checks whether the mouse is on the image, and not between the image and
    the border of the canvas. Returns :obj:`True` if it is on the image,
    :obj:`False` otherwise."""

    if self._pil_img is None:
      return False

    if abs(event.x -
           self._img_canvas.winfo_width() / 2) > self._pil_img.width / 2:
      return False
    if abs(event.y -
           self._img_canvas.winfo_height() / 2) > self._pil_img.height / 2:
      return False

    return True

  def _add_settings(self) -> None:
    """Adds the settings of the camera to the GUI."""

    self.log(logging.DEBUG, "Adding the camera settings to the interface")

    # First, sort the settings by type for a nicer display
    sort_sets = sorted(self._camera.settings.values(),
                       key=lambda setting: setting.type.__name__)

    for cam_set in sort_sets:
      if isinstance(cam_set, CameraBoolSetting):
        self._add_bool_setting(cam_set)
      elif isinstance(cam_set, CameraScaleSetting):
        self._add_slider_setting(cam_set)
      elif isinstance(cam_set, CameraChoiceSetting):
        self._add_choice_setting(cam_set)

  def _add_bool_setting(self, cam_set: CameraBoolSetting) -> None:
    """Adds a setting represented by a checkbutton."""

    self.log(logging.DEBUG, f"Adding the boolean setting {cam_set.name}")

    cam_set.tk_var = tk.BooleanVar(value=cam_set.value)
    cam_set.tk_obj = tk.Checkbutton(self._canvas_frame,
                                    text=cam_set.name,
                                    variable=cam_set.tk_var,
                                    command=self._auto_apply_bool_settings)

    cam_set.tk_obj.pack(anchor='w', side='top', expand=False, fill='none',
                        padx=5, pady=2)

  def _add_slider_setting(self, cam_set: CameraScaleSetting) -> None:
    """Adds a setting represented by a scale bar."""

    self.log(logging.DEBUG, f"Adding the slider setting {cam_set.name}")

    # The scale bar is slightly different if the setting type is int or float
    if cam_set.type == int:
      cam_set.tk_var = tk.IntVar(value=cam_set.value)
    else:
      cam_set.tk_var = tk.DoubleVar(value=cam_set.value)

    cam_set.tk_obj = tk.Scale(self._canvas_frame,
                              label=f'{cam_set.name} :',
                              variable=cam_set.tk_var,
                              resolution=cam_set.step,
                              orient='horizontal',
                              from_=cam_set.lowest,
                              to=cam_set.highest)

    cam_set.tk_obj.bind("<ButtonRelease-1>", self._auto_apply_scale_settings)

    cam_set.tk_obj.pack(anchor='center', side='top', expand=False,
                        fill='x', padx=5, pady=2)

  def _add_choice_setting(self, cam_set: CameraChoiceSetting) -> None:
    """Adds a setting represented by a list of radio buttons."""

    self.log(logging.DEBUG, f"Adding the choice setting {cam_set.name}")

    cam_set.tk_var = tk.StringVar(value=cam_set.value)
    label = tk.Label(self._canvas_frame, text=f'{cam_set.name} :')
    label.pack(anchor='w', side='top', expand=False, fill='none',
               padx=12, pady=2)

    for value in cam_set.choices:
      tk_obj = tk.Radiobutton(self._canvas_frame,
                              text=value,
                              variable=cam_set.tk_var,
                              value=value,
                              command=self._auto_apply_choice_settings)

      tk_obj.pack(anchor='w', side='top', expand=False,
                  fill='none', padx=5, pady=2)

      cam_set.tk_obj.append(tk_obj)

  def _set_variables(self) -> None:
    """Sets the text and numeric variables holding information about the
    display."""

    self.log(logging.DEBUG, "Setting the interface variables")

    # The FPS counter
    self._fps_var = tk.DoubleVar(value=0.)
    self._fps_txt = tk.StringVar(
        value=f'fps = {self._fps_var.get():.2f}\n(might be lower in this GUI '
              f'than actual)')

    # The variable for enabling or disabling the auto range
    self._auto_range = tk.BooleanVar(value=False)

    # The variable for enabling or disabling the auto apply
    self._auto_apply = tk.BooleanVar(value=False)

    # The minimum and maximum pixel value counters
    self._min_pixel = tk.IntVar(value=0)
    self._max_pixel = tk.IntVar(value=0)
    self._min_max_pix_txt = tk.StringVar(
      value=f'min: {self._min_pixel.get():d}, '
            f'max: {self._max_pixel.get():d}')

    # The number of detected bits counter
    self._nb_bits = tk.IntVar(value=0)
    self._bits_txt = tk.StringVar(
      value=f'Detected bits: {self._nb_bits.get():d}')

    # The display of the current zoom level
    self._zoom_level = tk.DoubleVar(value=100.0)
    self._zoom_txt = tk.StringVar(
      value=f'Zoom: {self._zoom_level.get():.1f}%')

    # The display of the current pixel position and value
    self._x_pos = tk.IntVar(value=0)
    self._y_pos = tk.IntVar(value=0)
    self._reticle_val = tk.IntVar(value=0)
    self._reticle_txt = tk.StringVar(value=f'X: {self._x_pos.get():d}, '
                                           f'Y: {self._y_pos.get():d}, '
                                           f'V: {self._reticle_val.get():d}')

  def _set_traces(self) -> None:
    """Sets the traces for automatically updating the display when a variable
    is modified."""

    self.log(logging.DEBUG, "Setting the interface traces")

    self._fps_var.trace_add('write', self._update_fps)

    self._min_pixel.trace_add('write', self._update_min_max)
    self._max_pixel.trace_add('write', self._update_min_max)

    self._nb_bits.trace_add('write', self._update_bits)

    self._zoom_level.trace_add('write', self._update_zoom)

    self._x_pos.trace_add('write', self._update_reticle)
    self._y_pos.trace_add('write', self._update_reticle)
    self._reticle_val.trace_add('write', self._update_reticle)

    self._auto_apply.trace_add('write', self._update_apply_settings)

  def _update_fps(self, _, __, ___) -> None:
    """Auto-update of the FPS display."""

    self._fps_txt.set(f'fps = {self._fps_var.get():.2f}\n'
                      f'(might be lower in this GUI than actual)')

  def _update_min_max(self, _, __, ___) -> None:
    """Auto-update of the minimum and maximum pixel values display."""

    self._min_max_pix_txt.set(f'min: {self._min_pixel.get():d}, '
                              f'max: {self._max_pixel.get():d}')

  def _update_bits(self, _, __, ___) -> None:
    """Auto-update of the number of detected bits display."""

    self._bits_txt.set(f'Detected bits: {self._nb_bits.get():d}')

  def _update_zoom(self, _, __, ___) -> None:
    """Auto-update of the current zoom level display."""

    self._zoom_txt.set(f'Zoom: {self._zoom_level.get():.1f}%')

  def _update_reticle(self, _, __, ___) -> None:
    """Auto-update of the current pixel position and value display."""

    self._reticle_txt.set(f'X: {self._x_pos.get():d}, '
                          f'Y: {self._y_pos.get():d}, '
                          f'V: {self._reticle_val.get():d}')

  def _update_apply_settings(self, _, __, ___) -> None:
    """Disable the Apply Settings button when Auto apply button is checked."""

    if self._auto_apply.get():
      self._update_button['state'] = 'disabled'
    else:
      self._update_button['state'] = 'normal'

  def _update_settings(self) -> None:
    """Tries to update the settings values upon clicking on the Apply Settings
    button, and checks that the settings have been correctly set."""

    for setting in self._camera.settings.values():
      # Applying all the settings that differ from the read value
      if setting.value != setting.tk_var.get():
        setting.value = setting.tk_var.get()

      # Reading the actual value of all the settings
      setting.tk_var.set(setting.value)

  def _auto_apply_scale_settings(self, _: tk.Event):
    """Applies the settings without clicking on the Apply Settings
     button when the Auto apply button is checked.

     The scale settings will be applied when the slicer is released.
     """

    if self._auto_apply.get():
      self._update_settings()

  def _auto_apply_bool_settings(self):
    """Applies the settings without clicking on the Apply Settings
     button when the Auto apply button is checked.

     The bool settings will be applied when the bool button is checked.
     """

    if self._auto_apply.get():
      self._update_settings()

  def _auto_apply_choice_settings(self):
    """Applies the settings without clicking on the Apply Settings
     button when the Auto apply button is checked.

     The choice settings will be applied when the choice button is checked.
     """

    if self._auto_apply.get():
      self._update_settings()

  def _cast_img(self, img: np.ndarray) -> None:
    """Casts the image to 8-bits as a greater precision is not required.

    May also interpolate the image to obtain a higher contrast, depending on
    the user's choice.
    """

    # First, convert BGR to RGB
    if len(img.shape) == 3:
      img = img[:, :, ::-1]

    # If the auto_range is set, adjusting the values to the range
    if self._auto_range.get():
      self.log(logging.DEBUG, "Applying auto range to the image")
      self._low_thresh, self._high_thresh = np.percentile(img, (3, 97))
      self._img = ((np.clip(img, self._low_thresh, self._high_thresh) -
                    self._low_thresh) * 255 /
                   (self._high_thresh - self._low_thresh)).astype('uint8')

      # The original image still needs to be saved as 8-bits
      bit_depth = np.ceil(np.log2(np.max(img) + 1))
      self._original_img = (img / 2 ** (bit_depth - 8)).astype('uint8')

    # Or if the image is not already 8 bits, casting to 8 bits
    elif img.dtype != np.uint8:
      self.log(logging.DEBUG, "Casting the image to 8 bits")
      bit_depth = np.ceil(np.log2(np.max(img) + 1))
      self._img = (img / 2 ** (bit_depth - 8)).astype('uint8')
      self._original_img = np.copy(self._img)

    # Else, the image is usable as is
    else:
      self._img = img
      self._original_img = np.copy(img)

    # Updating the information
    self._nb_bits.set(int(np.ceil(np.log2(np.max(img) + 1))))
    self._max_pixel.set(int(np.max(img)))
    self._min_pixel.set(int(np.min(img)))

  def _resize_img(self) -> None:
    """Resizes the received image so that it fits in the image canvas and
    complies with the chosen zoom level."""

    if self._img is None:
      return

    self.log(logging.DEBUG, "Resizing the image to fit in the window")

    # First, apply the current zoom level
    # The width and height values are inverted in NumPy
    img_height, img_width, *_ = self._img.shape
    y_min_pix = int(img_height * self._zoom_values.y_low)
    y_max_pix = int(img_height * self._zoom_values.y_high)
    x_min_pix = int(img_width * self._zoom_values.x_low)
    x_max_pix = int(img_width * self._zoom_values.x_high)
    zoomed_img = self._img[y_min_pix: y_max_pix, x_min_pix: x_max_pix]

    # Creating the pillow image from the zoomed numpy array
    pil_img = Image.fromarray(zoomed_img)

    # Resizing the image to make it fit in the image canvas
    img_canvas_width = self._img_canvas.winfo_width()
    img_canvas_height = self._img_canvas.winfo_height()

    zoomed_img_ratio = pil_img.width / pil_img.height
    img_label_ratio = img_canvas_width / img_canvas_height

    if zoomed_img_ratio >= img_label_ratio:
      new_width = img_canvas_width
      new_height = max(int(img_canvas_width / zoomed_img_ratio), 1)
    else:
      new_width = max(int(img_canvas_height * zoomed_img_ratio), 1)
      new_height = img_canvas_height

    self._pil_img = pil_img.resize((new_width, new_height))

  def _display_img(self) -> None:
    """Displays the image in the center of the image canvas."""

    if self._pil_img is None:
      return

    self.log(logging.DEBUG, "Displaying the image")

    self._image_tk = ImageTk.PhotoImage(self._pil_img)
    self._img_canvas.create_image(int(self._img_canvas.winfo_width() / 2),
                                  int(self._img_canvas.winfo_height() / 2),
                                  anchor='center', image=self._image_tk)

  def _on_img_resize(self, _: Optional[tk.Event] = None) -> None:
    """Resizes the image and updates the display when the zoom level has
    changed or the GUI has been resized."""

    self.log(logging.DEBUG, "The image canvas was resized")

    self._resize_img()
    self._display_img()
    self.update()

  def _calc_hist(self) -> None:
    """Calculates the histogram of the current image."""

    if self._original_img is None:
      return

    # Don't calculate histogram if a calculation is already running
    if self._processing_event.is_set():
      self.log(logging.DEBUG, "A calculation is running for the histogram, "
                              "not sending image for calculation")
      return

    # If no calculation is running, sending a new image for calculation
    else:
      # Reshaping the image before sending to the histogram process
      self.log(logging.DEBUG, "Preparing image for histogram calculation")
      hist_img = Image.fromarray(self._original_img)
      if hist_img.width > 320 or hist_img.height > 240:
        factor = min(320 / hist_img.width, 240 / hist_img.height)
        hist_img = hist_img.resize((max(int(hist_img.width * factor), 1),
                                    max(int(hist_img.height * factor), 1)))
      # The histogram is calculated on a grey level image
      if len(self._original_img.shape) == 3:
        hist_img = hist_img.convert('L')

      # Sending the image to the histogram process
      self.log(logging.DEBUG, "Sending image for histogram calculation")
      self._img_in.put_nowait((hist_img, self._auto_range.get(),
                               self._low_thresh, self._high_thresh))

    # Checking if a histogram is available for display
    while not self._img_out.empty():
      self._hist = self._img_out.get_nowait()
      self.log(logging.DEBUG, "Received histogram from histogram process")

  def _resize_hist(self) -> None:
    """Resizes the histogram image to make it fit in the GUI."""

    if self._hist is None:
      return

    self.log(logging.DEBUG, "Resizing the histogram to fit in the window")

    pil_hist = Image.fromarray(self._hist)
    hist_canvas_width = self._hist_canvas.winfo_width()
    hist_canvas_height = self._hist_canvas.winfo_height()

    self._pil_hist = pil_hist.resize((hist_canvas_width, hist_canvas_height))

  def _display_hist(self) -> None:
    """Displays the histogram image in the GUI."""

    if self._pil_hist is None:
      return

    self.log(logging.DEBUG, "Displaying the histogram")

    self._hist_tk = ImageTk.PhotoImage(self._pil_hist)
    self._hist_canvas.create_image(int(self._hist_canvas.winfo_width() / 2),
                                   int(self._hist_canvas.winfo_height() / 2),
                                   anchor='center', image=self._hist_tk)

  def _on_hist_resize(self, _: tk.Event) -> None:
    """Resizes the histogram and updates the display when the GUI has been
    resized."""

    self._resize_hist()
    self._display_hist()
    self.update()

  def _update_img(self) -> None:
    """Acquires an image from the camera, casts and resizes it, calculates its
    histogram, displays them and updates the image information."""

    self.log(logging.DEBUG, "Updating the image")

    ret = self._camera.get_image()

    # If no frame could be grabbed from the camera
    if ret is None:
      # If it's the first call, generate error image to initialize the window
      if not self._n_loops:
        self.log(logging.WARNING, "Could not get an image from the camera, "
                                  "displaying an error image instead")
        ret = None, np.array(Image.open(BytesIO(resource_string(
          'crappy', 'tool/data/no_image.png'))))
      # Otherwise, just pass
      else:
        self.log(logging.DEBUG, "No image returned by the camera")
        self.update()
        sleep(0.001)
        return

    self._n_loops += 1
    _, img = ret

    if img.dtype != self.dtype:
      self.dtype = img.dtype
    if self.shape != img.shape:
      self.shape = img.shape

    self._cast_img(img)
    self._resize_img()

    self._calc_hist()
    self._resize_hist()

    self._display_img()
    self._display_hist()

    self._update_pixel_value()

    self.update()



# coding: utf-8

from typing import Union, Optional, Iterable, Any
import logging

from .meta_block import Block
from ..inout import inout_dict, InOut, deprecated_inouts
from ..tool.ft232h import USBServer


class IOBlock(Block):
  """This Block is meant to drive :class:`~crappy.inout.InOut` objects. It can
  acquire data, and/or set commands. One IOBlock can only drive a single InOut.

  If it has incoming :class:`~crappy.links.Link`, it will set the commands 
  received over the labels given in ``cmd_labels`` by calling the 
  :meth:`~crappy.inout.InOut.set_cmd` method of the InOut. Additional commands 
  to set at the very beginning or the very end of the test can also be 
  specified.

  If it has outgoing :class:`~crappy.links.Link`, it will acquire data using 
  the :meth:`~crappy.inout.InOut.get_data` method of the InOut and send it
  downstream over the labels given in ``labels``. It is possible to trigger the
  acquisition using a predefined label.

  The ``streamer`` argument allows using the "streamer" mode of InOuts
  supporting it, instead of the regular acquisition mode. Finally, the
  ``make_zero_delay`` argument allows offsetting the acquired values to zero at
  the beginning of the test. Refer to the documentation of each argument for a
  more detailed description.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               name: str,
               labels: Optional[Union[str, Iterable[str]]] = None,
               cmd_labels: Optional[Union[str, Iterable[str]]] = None,
               trigger_label: Optional[str] = None,
               streamer: bool = False,
               initial_cmd: Optional[Union[Any, Iterable[Any]]] = None,
               exit_cmd: Optional[Union[Any, Iterable[Any]]] = None,
               make_zero_delay: Optional[float] = None,
               ft232h_ser_num: Optional[str] = None,
               spam: bool = False,
               freq: Optional[float] = 200,
               display_freq: bool = False,
               debug: Optional[bool] = False,
               **kwargs) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      name: The name of the :class:`~crappy.inout.InOut` class to instantiate.
      labels: An iterable (e.g. a :obj:`list` or a :obj:`tuple`) containing the
        output labels for InOuts that acquire data. They correspond to the
        values returned by the InOut's :meth:`~crappy.inout.InOut.get_data`
        method, so there should be as many labels as returned values, and given
        in the appropriate order. The first label must always be the time
        label, preferably called ``'t(s)'``. This argument can be omitted if
        :meth:`~crappy.inout.InOut.get_data` returns a :obj:`dict`. Ignored if
        the Block has no output Link.
      cmd_labels: An iterable (e.g. a :obj:`list` or a :obj:`tuple`) containing
        the labels considered as inputs of this Block, for InOuts that set
        commands. The values received from these labels will be passed to the
        InOut's :meth:`~crappy.inout.InOut.set_cmd` method, in the same order
        as the labels are given. Usually, time is not part of the
        ``cmd_labels``. Ignored if the Block has no input Link.
      trigger_label: If given, the Block will only read data whenever a value
        is received on this label (can be any value). Ignored if the Block has
        no output Link. A trigger label can also be a cmd label.
      streamer: If :obj:`False`, the :meth:`~crappy.inout.InOut.get_data`
        method of the InOut is called for acquiring data, else it is the
        :meth:`~crappy.inout.InOut.get_stream` method. Refer to the
        documentation of these methods for more information.
      initial_cmd: An initial command for the InOut, set during
        :meth:`prepare`. If given, there must be as many values as in
        ``cmd_labels``. Must be given as an iterable (e.g. a :obj:`list` or a
        :obj:`tuple`).
      exit_cmd: A final command for the InOut, set during :meth:`finish`. If
        given, there must be as many values as in ``cmd_labels``. Must be given
        as an iterable (e.g. a :obj:`list` or a :obj:`tuple`).
      make_zero_delay: If set, will acquire data before the beginning of the
        test and use it to offset all the labels to zero. The data will be
        acquired during the given number of seconds. Ignored if the Block has
        no output Links. Does not work for InOuts that acquire values other
        than numbers (:obj:`str` for example).
      spam: If :obj:`False`, the Block will call
        :meth:`~crappy.inout.InOut.set_cmd` on the InOut object only if the
        current command is different from the previous. Otherwise, it will call
        the method each time a command is received.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block while running.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
      **kwargs: The arguments to be passed to the :class:`~crappy.inout.InOut`.

    .. versionchanged:: 1.5.10 renamed *trigger* argument to *trigger_label*
    .. versionchanged:: 1.5.10 renamed *exit_values* argument to *exit_cmd*
    .. versionadded:: 1.5.10 *make_zero_delay* argument
    .. versionadded 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    self._device: Optional[InOut] = None
    self._ft232h_args = None
    self._read: bool = False
    self._write: bool = False

    super().__init__()
    self.niceness = -10
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    # The label argument can be omitted for streaming
    if labels is None and streamer:
      self.labels = ['t(s)', 'stream']
    # Forcing the labels into a list
    elif labels is not None and isinstance(labels, str):
      self.labels = [labels]
    elif labels is not None:
      self.labels = list(labels)
    else:
      self.labels = None

    # Forcing the cmd_labels into a list or None
    if cmd_labels is not None and isinstance(cmd_labels, str):
      self._cmd_labels = [cmd_labels]
    elif cmd_labels is not None:
      self._cmd_labels = list(cmd_labels)
    else:
      self._cmd_labels = None

    # Forcing the initial_cmd into a list
    if initial_cmd is not None and isinstance(initial_cmd, str):
      self._initial_cmd = [initial_cmd]
    elif initial_cmd is not None:
      self._initial_cmd = list(initial_cmd)
    else:
      self._initial_cmd = None

    # Forcing the exit_cmd into a list
    if exit_cmd is not None and isinstance(exit_cmd, str):
      self._exit_cmd = [exit_cmd]
    elif exit_cmd is not None:
      self._exit_cmd = list(exit_cmd)
    else:
      self._exit_cmd = None

    # Checking that the initial_cmd and exit_cmd length are consistent
    if self._cmd_labels is not None:
      if self._initial_cmd is not None \
          and len(self._initial_cmd) != len(self._cmd_labels):
        raise ValueError("There should be as many values in initial_cmd as "
                         "there are in cmd_labels !")
      if self._exit_cmd is not None \
          and len(self._exit_cmd) != len(self._cmd_labels):
        raise ValueError("There should be as many values in exit_cmd as "
                         "there are in cmd_labels !")

    self._trig_label = trigger_label

    # Checking for deprecated names
    if name in deprecated_inouts:
      raise NotImplementedError(
          f"The {name} InOut was deprecated in version 2.0.0, and renamed "
          f"to {deprecated_inouts[name]} ! Please update your code "
          f"accordingly and check the documentation for more information")

    # Checking that all the given actuators are valid
    if name not in inout_dict:
      possible = ', '.join(sorted(inout_dict.keys()))
      raise ValueError(f"Unknown InOut type : {name} ! "
                       f"The possible types are : {possible}")

    self._io_name = name
    self._inout_kwargs = kwargs

    self._streamer = streamer
    self._spam = spam
    self._make_zero_delay = make_zero_delay

    self._stream_started = False
    self._last_cmd = None
    self._prev_values = dict()

    # Checking whether the InOut communicates through an FT232H
    if inout_dict[self._io_name].ft232h:
      self._ft232h_args = USBServer.register(ft232h_ser_num)

  def prepare(self) -> None:
    """Checks the consistency of the Link layout, opens the InOut and sets the
    initial command if required.

    This method mainly calls the :meth:`~crappy.inout.InOut.open` method of the
    driven InOut.
    """

    # Instantiating the device in a regular way
    if self._ft232h_args is None:
      self._device = inout_dict[self._io_name](**self._inout_kwargs)
    # Instantiating the device and the connection to the FT232H
    else:
      self.log(logging.INFO, "The InOut to open communicates over an FT232H")
      self._device = inout_dict[self._io_name](**self._inout_kwargs,
                                               _ft232h_args=self._ft232h_args)

    # Checking that the block has inputs or outputs
    if not self.inputs and not self.outputs:
      raise IOError('Error ! The IOBlock is neither an input nor an output !')

    # cmd_labels must be defined when the block has inputs
    if self.inputs and self._cmd_labels is None and self._trig_label is None:
      raise ValueError('Error ! The IOBlock has incoming links but no '
                       'cmd_labels have been given !')

    self._read = bool(self.outputs)
    self._write = bool(self._cmd_labels)

    # Now opening the device
    self.log(logging.INFO, f"Opening the {type(self._device).__name__} InOut")
    self._device.open()
    self.log(logging.INFO, f"{type(self._device).__name__} InOut opened")

    # Acquiring data for offsetting the output
    if self._read and self._make_zero_delay is not None:
      self.log(logging.INFO, f"Performing offsetting on the "
                             f"{type(self._device).__name__} InOut")
      self._device.make_zero(self._make_zero_delay)

    # Writing the first command before the beginning of the test if required
    if self._write and self._initial_cmd is not None:
      self.log(logging.INFO, f"Sending the initial command to the "
                             f"{type(self._device).__name__} InOut")
      self._device.set_cmd(*self._initial_cmd)
      self._last_cmd = self._initial_cmd
      self._prev_values.update(zip(self._cmd_labels, self._initial_cmd))

  def loop(self) -> None:
    """Reads data from the InOut and/or sets the received commands.

    Data is read from the InOut **only** if this Block has outgoing Links. If
    the ``trigger_label`` is given, data is read only if a trigger is received
    over the given trigger label.

    A command is set on the InOut **only** if this Block has incoming Links,
    and if data is received over these Links. Depending on the value of the
    ``spam`` argument, a command might not be set if it is similar to the
    previous one.

    The data is read from the InOut either by calling its
    :meth:`~crappy.inout.InOut.return_data` or its
    :meth:`~crappy.inout.InOut.return_stream` method, depending if the
    ``streamer`` argument is :obj:`True` of :obj:`False`. The commands are
    always set by calling the :meth:`~crappy.inout.InOut.set_cmd` method.
    """

    # Receiving all the latest data waiting in the links
    data = self.recv_last_data(fill_missing=False)

    # Reading data from the device if there's no trig_label or if data has been
    # received on this trig_label
    if self._read:
      if self._trig_label is None:
        self._read_data()
      elif self._trig_label in data:
        self.log(logging.DEBUG, "Software trigger signal received")
        self._read_data()

    # If no data was received, there's nothing to write
    if not data:
      return

    if self._write:
      # The missing values are completed here, because the trig label must not
      # be artificially created
      self._prev_values.update(data)
      data.update(self._prev_values)

      # Keeping only the labels in cmd_labels
      data = {key: val for key, val in data.items() if key in self._cmd_labels}

      # If not all cmd_labels have a value, returning without calling set_cmd
      if len(data) != len(self._cmd_labels):
        self.log(logging.WARNING, f"Not enough values received in the "
                                  f"{type(self._device).__name__} InOut to"
                                  f" set the cmd, cmd not set !")
        return

      # Grouping the command values in a list before passing them to set_cmd
      cmd = [data[label] for label in self._cmd_labels]

      # Setting the command if it's different from the previous or spam is True
      if cmd != self._last_cmd or self._spam:
        self.log(logging.DEBUG, f"Writing the command {cmd} to the "
                                f"{type(self._device).__name__} InOut")
        self._device.set_cmd(*cmd)
        self._last_cmd = cmd

  def finish(self) -> None:
    """Stops the stream, sets the exit command if necessary, and closes the
    InOut.

    This method mainly calls the :meth:`~crappy.inout.InOut.close` method of
    the driven InOut.
    """

    # Stopping the stream
    if self._streamer and self._device is not None:
      self.log(logging.INFO, f"Stopping stream on the "
                             f"{type(self._device).__name__} InOut")
      self._device.stop_stream()

    # Setting the exit command
    if self._write and self._exit_cmd is not None and self._device is not None:
      self.log(logging.INFO, f"Sending the exit command to the "
                             f"{type(self._device).__name__} InOut")
      self._device.set_cmd(*self._exit_cmd)

    # Closing the device
    if self._device is not None:
      self.log(logging.INFO, f"Closing the {type(self._device).__name__} "
                             f"InOut")
      self._device.close()
      self.log(logging.INFO, f"{type(self._device).__name__} InOut closed")

  def _read_data(self) -> None:
    """Reads the data or the stream, offsets the timestamp and sends the data
    to downstream Blocks."""

    if self._streamer:
      # Starting the stream if needed
      if not self._stream_started:
        self.log(logging.INFO, f"Starting stream on the "
                               f"{type(self._device).__name__} InOut")
        self._device.start_stream()
        self._stream_started = True
      # Actually getting the stream
      data = self._device.return_stream()
    else:
      # Regular reading of data
      data = self._device.return_data()

    self.log(logging.DEBUG, f"Read values {data} from the "
                            f"{type(self._device).__name__} InOut")

    if data is None:
      return

    # Making time relative to the beginning of the test
    if isinstance(data, dict) and 't(s)' in data:
      data['t(s)'] -= self.t0
    else:
      data[0] -= self.t0

    self.send(data)



# coding: utf-8

from time import time
from typing import Optional, List
import logging

from ..meta_inout import InOut
from ...tool.ft232h import FT232HServer as FT232H, I2CMessage, USBArgsType

mprls_status_bits = {'busy': 0x20,
                     'memory error': 0x04,
                     'math saturation': 0x01}


class MPRLSFT232H(InOut):
  """This class can read values from an MPRLS pressure sensor through an
  FT232H.

  It is similar to the :class:`~crappy.inout.MPRLS` class, except this class is
  specific for use with an :class:`~crappy.tool.ft232h.FT232H` USB to I2C
  converter.

  It communicates over I2C with the sensor.

  .. versionadded:: 2.0.0
  """

  ft232h = True

  def __init__(self,
               eoc_pin: Optional[str] = None,
               device_address: int = 0x18,
               _ft232h_args: USBArgsType = tuple()) -> None:
    """Initializes the parent class and opens the I2C bus.

    Args:
      eoc_pin: Optionally, reads the end of conversion signal from the polarity
        of a GPIO rather than from an I2C register. Speeds up the reading and
        decreases the traffic on the bus, but requires one extra wire. Give the
        name of the GPIO in the format `Dx` or `Cx`.
      device_address: The I2C address of the MPRLS. The address of the devices
        sold by Adafruit is `0x18`, but other suppliers may sell it with
        another address.
      _ft232h_args: This argument is meant for internal use only and should not
        be provided by the user. It contains the information necessary for
        setting up the FT232H.
    """

    self._bus = None

    super().__init__()

    (block_index, block_lock, command_file, answer_file, shared_lock,
     current_block) = _ft232h_args

    self._bus = FT232H(mode='I2C',
                       block_index=block_index,
                       current_block=current_block,
                       command_file=command_file,
                       answer_file=answer_file,
                       block_lock=block_lock,
                       shared_lock=shared_lock)

    if not isinstance(device_address, int):
      raise TypeError("device_address should be an integer.")
    self._address = device_address

    if eoc_pin is not None and not isinstance(eoc_pin, str):
      raise TypeError('eoc_pin should be a string when using the ft232h '
                      'backend !')
    self._eoc_pin = eoc_pin

    self._i2c_msg = None

  def open(self) -> None:
    """Opens the I2C bus."""

    self._i2c_msg = I2CMessage

  def get_data(self) -> List[float]:
    """Reads the pressure value.

    Returns:
      The timestamp and the pressure value in hPa.
    """

    # Starting conversion
    self.log(logging.DEBUG, f"Writing {0xAA, 0x00, 0x00} to the address "
                            f"{self._address}")
    self._bus.i2c_rdwr(self._i2c_msg.write(self._address,
                                           [0xAA, 0x00, 0x00]))
    # Waiting for conversion to complete
    t0 = time()
    while not self._data_available():
      if time() - t0 > 0.1:
        raise TimeoutError('Waited too long for data to be ready')

    # Reading conversion result
    read = self._i2c_msg.read(self._address, 4)
    self._bus.i2c_rdwr(read)
    out = list(read)
    self.log(logging.DEBUG, f"Read {out} from address {self._address}")
    # Checking if anu error occurred
    if out[0] & mprls_status_bits['memory error']:
      raise RuntimeError("A memory error occurred on the MPRLS")
    elif out[0] & mprls_status_bits['math saturation']:
      raise RuntimeError("A math saturation error occurred on the MPRLS")
    # Extracting conversion result as an integer
    ret = (out[1] << 16) | (out[2] << 8) | out[3]
    # Converting to hPa
    pres = 68.947572932 * (ret - 0x19999A) * 25 / (0xE66666 - 0x19999A)

    return [time(), pres]

  def close(self) -> None:
    """Closes the I2C bus."""

    if self._bus is not None:
      self.log(logging.INFO, "Closing the I2C connection to the MPRLS")
      self._bus.close()

  def _data_available(self) -> bool:
    """Returns :obj:`True` if data is available, :obj:`False` otherwise."""

    # EOC signal from the I2C communication
    if self._eoc_pin is None:
      wait = self._i2c_msg.read(self._address, 1)
      self._bus.i2c_rdwr(wait)
      return not list(wait)[0] & mprls_status_bits['busy']
    # EOC signal from a GPIO
    else:
      return bool(self._bus.get_gpio(self._eoc_pin))



# coding: utf-8

from typing import Optional, Callable, Union, Tuple, Iterable
import numpy as np
from pathlib import Path

from .camera_processes import DISCorrelProcess
from .camera import Camera
from ..tool.camera_config import DISCorrelConfig, Box
from .._global import CameraConfigError


class DISCorrel(Camera):
  """This Block can perform Dense Inverse Search on a sub-frame (patch) of
  images acquired by a :class:`~crappy.camera.Camera` object, and project the
  result on various fields.

  It is mostly used for computing the displacement and the strain over the
  given patch, but other fields are also available. refer to the ``fields`` and
  ``labels`` arguments for more details. It relies on OpenCV's DISFlow
  algorithm, and offers the possibility to adjust many of its settings.

  This Block takes no input :class:`~crappy.links.Link` in a majority of
  situations, and outputs the results of image correlation. It is a subclass of
  the :class:`~crappy.blocks.Camera` Block, and inherits of all its features.
  That includes the possibility to record and to display images in real-time,
  simultaneously to the image acquisition and processing. Refer to the
  documentation of the Camera Block for more information on these features.

  This Block is very similar to the :class:`GPUCorrel` Block, except this
  latter uses GPU-acceleration to perform the image correlation and does not
  use DISFlow. The :class:`~crappy.blocks.DICVE` Block also relies on image
  correlation for computing the displacement and strain on images, but it
  tracks multiple patches and uses video-extensometry.

  Similar to the :class:`~crappy.tool.camera_config.CameraConfig` window that
  can be displayed by the Camera Block, this Block can display a
  :class:`~crappy.tool.camera_config.DISCorrelConfig` window before the test
  starts. Here, the user can also select the patch to track if it was not
  already specified as an argument.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               camera: str,
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               config: bool = True,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               software_trig_label: Optional[str] = None,
               display_freq: bool = False,
               freq: Optional[float] = 200,
               debug: Optional[bool] = False,
               save_images: bool = False,
               img_extension: str = "tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               img_shape: Optional[Tuple[int, int]] = None,
               img_dtype: Optional[str] = None,
               patch: Optional[Tuple[int, int, int, int]] = None,
               fields: Union[str, Iterable[str]] = None,
               labels: Union[str, Iterable[str]] = None,
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               iterations: int = 1,
               gradient_iterations: int = 10,
               init: bool = True,
               patch_size: int = 8,
               patch_stride: int = 3,
               residual: bool = False,
               **kwargs) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      camera: The name of the :class:`~crappy.camera.Camera` object to use for
        acquiring the images. Arguments can be passed to this Camera as
        ``kwargs`` of this Block. This argument is ignored if the
        ``image_generator`` argument is provided.
      transform: A callable taking an image as an argument, and returning a
        transformed image as an output. Allows applying a post-processing
        operation to the acquired images. This is done right after the
        acquisition, so the original image is permanently lost and only the
        transformed image is displayed and/or saved and/or further processed.
        The transform operation is not parallelized, so it might negatively
        affect the acquisition framerate if it is too heavy.
      config: If :obj:`True`, a
        :class:`~crappy.tool.camera_config.DISCorrelConfig` window is displayed
        before the test starts. There, the user can interactively adjust the
        different
        :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting`
        available for the selected :class:`~crappy.camera.Camera`, visualize
        the acquired images, and select the patch to track if it hasn't
        been given in the ``patch`` argument. The test starts when closing
        the configuration window. If not enabled, the ``img_dtype``,
        ``img_shape`` and ``patch`` arguments must be provided.
      display_images: If :obj:`True`, displays the acquired images in a
        dedicated window, using the backend given in ``displayer_backend`` and
        at the frequency specified in ``displayer_framerate``. This option
        should be considered as a debug or basic follow-up feature, it is not
        intended to be very fast nor to display high-quality images. The
        maximum resolution of the displayed images in `640x480`, the images
        might be downscaled to fit in this format. In addition to the acquired
        frames, the tracked patch is also displayed on the image as an
        overlay.
      displayer_backend: The backend to use for displaying the images. Can be
        either ``'cv2'`` or ``'mpl'``, to use respectively :mod:`cv2` (OpenCV)
        or :mod:`matplotlib`. ``'cv2'`` usually allows achieving a higher
        display frequency. Ignored if ``display_images`` is :obj:`False`. If
        not given and ``display_images`` is :obj:`True`, ``'cv2'`` is tried
        first and ``'mpl'`` second, and the first available one is used.
      displayer_framerate: The maximum update frequency of the image displayer,
        as an :obj:`int`. This value usually lies between 5 and 30Hz, the
        default is 5. The achieved update frequency might be lower than
        requested. Ignored if ``display_images`` is :obj:`False`.
      software_trig_label: The name of a label used as a software trigger for
        the :class:`~crappy.camera.Camera`. If given, images will only be
        acquired when receiving data over this label. The received value does
        not matter. This software trigger is not meant to be very precise, it
        is recommended not to rely on it for a trigger frequency greater than
        10Hz, in which case a hardware trigger should be preferred if available
        on the camera.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      save_images: If :obj:`True`, the acquired images are saved to the folder
        specified in ``save_folder``, in the format specified in
        ``img_extension``, using the backend specified in ``save_backend``, and
        at the frequency specified in ``save_period``. Each image is saved with
        the name : ``<frame_nr>_<timestamp>.<extension>``, and can thus easily
        be identified. Along with the images, a ``metadata.csv`` file records
        the metadata of all the saved images. This metadata is either the one
        returned by the :meth:`~crappy.camera.Camera.get_image` method of the
        :class:`~crappy.camera.Camera` object, or the default one generated in
        the :meth:`~crappy.blocks.Camera.loop` method of the
        :class:`~crappy.blocks.Camera` Block. Depending on the framerate of the
        camera and the performance of the computer, it is not guaranteed that
        all the acquired images will be recorded.
      img_extension: The file extension for the recorded images, as a
        :obj:`str` and without the dot. Common file extensions include `tiff`,
        `png`, `jpg`, etc. Depending on the used ``save_backend``, some
        extensions might not be available. It is currently not possible to
        customize the save parameters further than choosing the file extension.
        Ignored if ``save_images`` is :obj:`False`.
      save_folder: Path to the folder where to save the images, either as a
        :obj:`str` or as a :obj:`pathlib.Path`. Can be an absolute or a
        relative path, pointing to a folder. If the folder does not exist, it
        will be created (if the user has permission). If the given folder
        already contains a ``metadata.csv`` file (and thus likely images from
        Crappy), images are saved to another folder with the same name except
        a suffix is appended. Ignored if ``save_images`` is :obj:`False`. If
        not provided and ``save_images`` is :obj:`True`, the images are saved
        to the folder ``Crappy_images``, created next to the running script.
      save_period: Must be given as an :obj:`int`. Only one out of that number
        images at most will be saved. Allows to have a known periodicity in
        case the framerate is too high to record all the images. Or simply to
        reduce the number of recorded images if saving them all is not needed.
        Ignored if ``save_images`` is :obj:`False`.
      save_backend: If ``save_images`` is :obj:`True`, the backend to use for
        recording the images. It should be one of:
        ::

          'sitk', 'cv2', 'pil', 'npy'

        They correspond to the modules :mod:`SimpleITK`, :mod:`cv2` (OpenCV),
        :mod:`PIL` (Pillow Fork), and :mod:`numpy`. Note that the ``'npy'``
        backend saves the images as raw :obj:`numpy.array`, and thus ignores
        the ``img_extension`` argument. Depending on the machine, some backends
        may be faster or slower. For using each backend, the corresponding
        Python must of course be installed. If not provided and ``save_images``
        is :obj:`True`, the backends are tried in the same order as given above
        and the first available one is used. ``'npy'`` is always available.
      image_generator: A callable taking two :obj:`float` as arguments and
        returning an image as a :obj:`numpy.array`. **This argument is intended
        for use in the examples of Crappy, to apply an artificial strain on a
        base image. Most users should ignore it.** When given, the ``camera``
        argument is ignored and the images are acquired from the generator. To
        apply a strain on the image, strain values (in `%`) should be sent to
        the Camera Block over the labels ``'Exx(%)'`` and ``'Eyy(%)'``.
      img_shape: The shape of the images returned by the
        :class:`~crappy.camera.Camera` object as a :obj:`tuple` of :obj:`int`.
        It should correspond to the value returned by :obj:`numpy.shape`.
        **This argument is mandatory in case** ``config`` **is** :obj:`False`.
        It is otherwise ignored.
      img_dtype: The `dtype` of the images returned by the
        :class:`~crappy.camera.Camera` object, as a :obj:`str`. It should
        correspond to a valid data type in :mod:`numpy`, e.g. ``'uint8'``.
        **This argument is mandatory in case** ``config`` **is** :obj:`False`.
        It is otherwise ignored.
      patch: The coordinates of the patch to track, as a :obj:`tuple` of
        exactly 4 :obj:`int`. These integers correspond to the `y` position of
        the top-left corner of the patch, the `x` position of the top-left
        corner of the patch, the height of the patch, and the width of the
        patch. Only one patch can be tracked. This argument must be provided if
        ``config`` is :obj:`False`.
      fields: The several fields to calculate on the acquired images. They
        should be given as an iterable containing :obj:`str`. Each string
        represents one field to calculate, so the more fields are given the
        heavier the computation is. The possible fields are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

        For each field, one single value is computed, corresponding to the
        norm of the field values over the patch area. If not provided, the
        default fields are ``'x', 'y', 'exx', 'eyy'``.
      labels: The labels to use for sending data to downstream Blocks. If not
        given, the default labels are
        ``'t(s)', 'meta', 'x(pix)', 'y(pix)', 'Exx(%)', 'Eyy(%)'``. They carry
        for each image its timestamp, a :obj:`dict` containing its metadata,
        and then for each field the computed value as a :obj:`float`. These
        default labels are compatible with the default fields, but must be
        changed if the number of fields changes. When setting this argument,
        make sure to give at least 2 labels for the time and the metadata, and
        one label per field. The ``'res'`` label containing the residuals if
        ``residual`` is :obj:`True` should not be included here, it will be
        automatically added.
      alpha: Weight of the smoothness term in DISFlow, as a :obj:`float`.
      delta: Weight of the color constancy term in DISFlow, as a :obj:`float`.
      gamma: Weight of the gradient constancy term in DISFlow , as a
        :obj:`float`.
      finest_scale: Finest level of the Gaussian pyramid on which the flow is
        computed in DISFlow, as an :obj:`int`. Zero level corresponds to the
        original image resolution. The final flow is obtained by bilinear
        upscaling.
      iterations: The number of fixed point iterations of variational
        refinement per scale in DISFlow, as an :obj:`int`. Set to zero to
        disable variational refinement completely. Higher values will typically
        result in more smooth and high-quality flow.
      gradient_iterations: The maximum number of gradient descent iterations in
        the patch inverse search stage in DISFlow, as an :obj:`int`. Higher
        values may improve the quality.
      init: If :obj:`True`, the last calculated optical flow is used for
        initializing the calculation of the next one.
      patch_size: The size of an image patch for matching in DISFlow, in pixels
        as an :obj:`int`.
      patch_stride: The stride between two neighbor patches in DISFlow, in
        pixels as an :obj:`int`. Must be less than the ``patch_size``. Lower
        values correspond to higher flow quality.
      residual: If :obj:`True`, the residuals of the optical flow calculation
        are computed for each image. They are then returned under the ``'res'``
        label, that should not be included in the given labels. This option is
        mainly intended as a debug feature, to monitor the quality of the
        image correlation.
      **kwargs: Any additional argument will be passed to the
        :class:`~crappy.camera.Camera` object, and used as a kwarg to its
        :meth:`~crappy.camera.Camera.open` method.

    .. versionadded:: 1.5.10 
       *transform*, *config*, *displayer_backend*, *displayer_framerate*, 
       *verbose*, *freq*, *save_images*, *img_name*, *save_folder*, 
       *save_period*, *save_backend* and *image_generator* arguments
    .. versionchanged:: 1.5.10 
       renamed *gditerations* argument to *gradient_iterations*
    .. versionchanged:: 1.5.10 
       renamed *show_image* argument to *display_images*
    .. versionremoved:: 1.5.10 *residual_full* argument
    .. versionadded:: 2.0.0
       *debug*, *patch*, *software_trig_label*, *img_extension*, *img_size* and
       *img_dtype* arguments
    .. versionremoved:: 2.0.0 *img_name* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    if not config and patch is None:
      raise ValueError("If the config window is disabled, the patch must be "
                       "provided !")

    super().__init__(camera=camera,
                     transform=transform,
                     config=config,
                     display_images=display_images,
                     displayer_backend=displayer_backend,
                     displayer_framerate=displayer_framerate,
                     software_trig_label=software_trig_label,
                     display_freq=display_freq,
                     freq=freq,
                     debug=debug,
                     save_images=save_images,
                     img_extension=img_extension,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     img_shape=img_shape,
                     img_dtype=img_dtype,
                     **kwargs)

    # Forcing the fields into a list
    if fields is None:
      fields = ["x", "y", "exx", "eyy"]
    elif isinstance(fields, str):
      fields = [fields]
    else:
      fields = list(fields)

    # Forcing the labels into a list
    if labels is None:
      self.labels = ['t(s)', 'meta', 'x(pix)', 'y(pix)', 'Exx(%)', 'Eyy(%)']
    elif isinstance(labels, str):
      self.labels = [labels]
    else:
      self.labels = list(labels)

    # Adding the residuals if required
    if residual and labels is None:
      self.labels.append('res')

    self._patch_int = patch
    self._patch: Optional[Box] = None

    # Making sure a coherent number of labels and fields was given
    if 2 + len(fields) + int(residual) != len(self.labels):
      raise ValueError(
        "The number of fields is inconsistent with the number "
        "of labels !\nMake sure that the time label was given")

    # These arguments are for the DISCorrelProcess
    self._fields = fields
    self._alpha = alpha
    self._delta = delta
    self._gamma = gamma
    self._finest_scale = finest_scale
    self._init = init
    self._iterations = iterations
    self._gradient_iterations = gradient_iterations
    self._patch_size = patch_size
    self._patch_stride = patch_stride
    self._residual = residual

  def prepare(self) -> None:
    """This method mostly calls the :meth:`~crappy.blocks.Camera.prepare`
    method of the parent class.

    In addition to that it instantiates the
    :class:`~crappy.blocks.camera_processes.DISCorrelProcess` object that
    performs the image correlation and the tracking.

    .. versionchanged:: 1.5.5 now accepting args and kwargs
    .. versionchanged:: 1.5.10 not accepting arguments anymore
    """

    # Instantiating the Box containing the patch to track
    if self._patch_int is not None:
      self._patch = Box(x_start=self._patch_int[1],
                        x_end=self._patch_int[1] + self._patch_int[3],
                        y_start=self._patch_int[0],
                        y_end=self._patch_int[0] + self._patch_int[2])
    else:
      self._patch = Box()

    # Instantiating the DISCorrelProcess
    self.process_proc = DISCorrelProcess(
        patch=self._patch,
        fields=self._fields,
        alpha=self._alpha,
        delta=self._delta,
        gamma=self._gamma,
        finest_scale=self._finest_scale,
        init=self._init,
        iterations=self._iterations,
        gradient_iterations=self._gradient_iterations,
        patch_size=self._patch_size,
        patch_stride=self._patch_stride,
        residual=self._residual)

    super().prepare()

  def _configure(self) -> None:
    """This method should instantiate and start the
    :class:`~crappy.tool.camera_config.DISCorrelConfig` window for configuring
    the :class:`~crappy.camera.Camera` object.

    It should also handle the case when an exception is raised in the
    configuration window.
    """

    config = None

    # Instantiating and starting the configuration window
    try:
      config = DISCorrelConfig(self._camera, self._log_queue, self._log_level,
                               self.freq, self._patch)
      config.main()

    # If an exception is raised in the config window, closing it before raising
    except (Exception,) as exc:
      self._logger.exception("Caught exception in the configuration window !",
                             exc_info=exc)
      if config is not None:
        config.stop()
      raise CameraConfigError

    # Getting the image dtype and shape for setting the shared Array
    if config.shape is not None:
      self._img_shape = config.shape
    if config.dtype is not None:
      self._img_dtype = config.dtype



# coding: utf-8

import logging
from typing import Optional

from .meta_block import Block
from .._global import OptionalModule

try:
  import tkinter as tk
except (ModuleNotFoundError, ImportError):
  tk = OptionalModule("tkinter")


class StopButton(Block):
  """This Block allows the user to stop the current Crapy script by clicking on
  a button in a GUI.

  Along with the :class:`~crappy.blocks.StopBlock`, it allows to stop a test in
  a clean way without resorting to CTRL+C.

  .. versionadded 2.0.0
  """

  def __init__(self,
               freq: Optional[float] = 50,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    """

    self._root: Optional[tk.Tk] = None

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    self._label = None
    self._button = None

  def prepare(self) -> None:
    """Creates the graphical interface and sets its layout and callbacks."""

    self.log(logging.INFO, "Creating the GUI")

    self._root = tk.Tk()
    self._root.title("Stop Button Block")
    self._root.resizable(False, False)

    self._label = tk.Label(self._root, text="Click button to stop test")
    self._label.pack(padx=7, pady=7)

    self._button = tk.Button(self._root,
                             text='STOP',
                             command=self._clicked)
    self._button.pack(padx=25, pady=7)

    self._root.update()

  def loop(self) -> None:
    """Updates the interface at each loop."""

    try:
      self._root.update()
      self.log(logging.DEBUG, "GUI updated")
    except tk.TclError:
      return

  def finish(self) -> None:
    """Closes the interface window."""

    try:
      if self._root is not None:
        self.log(logging.INFO, "Closing the GUI")
        self._root.destroy()
    except tk.TclError:
      pass

  def _clicked(self) -> None:
    """When the stop button is clicked, stops the test."""

    self.log(logging.DEBUG, "Button clicked in the GUI")
    self.log(logging.WARNING, "Stop button clicked, stopping the script !")
    self.stop()



# coding: utf-8

from typing import Optional
from tkinter.messagebox import showerror
import tkinter as tk
import numpy as np
from io import BytesIO
from pkg_resources import resource_string
from time import sleep
import logging
from multiprocessing.queues import Queue

from .camera_config_boxes import CameraConfigBoxes
from .config_tools import Box, SpotsBoxes
from ...camera.meta_camera import Camera
from ...camera.meta_camera.camera_setting import CameraScaleSetting
from ..._global import OptionalModule

try:
  from PIL import Image
except (ModuleNotFoundError, ImportError):
  Image = OptionalModule("pillow")


class DICVEConfig(CameraConfigBoxes):
  """Class similar to :class:`~crappy.tool.camera_config.CameraConfig` but also 
  displaying the bounding boxes of the regions defined as patches.
  
  It relies on the :class:`~crappy.tool.camera_config.config_tools.Box` and 
  :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` tools. It is
  meant to be used for configuring the :class:`~crappy.blocks.DICVE` Block.
  
  .. versionadded:: 1.5.10
  .. versionchanged:: 2.0.0 renamed from DISVE_config to DICVEConfig
  """

  def __init__(self,
               camera: Camera,
               log_queue: Queue,
               log_level: Optional[int],
               max_freq: Optional[float],
               patches: SpotsBoxes) -> None:
    """Sets the patches and initializes the parent class.

    Args:
      camera: The :class:`~crappy.camera.Camera` object in charge of acquiring 
        the images.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to 
        the main :obj:`~logging.Logger`, only used in Windows.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.
      max_freq: The maximum frequency this window is allowed to loop at. It is
        simply the ``freq`` attribute of the :class:`~crappy.blocks.Camera`
        Block.
      patches: An instance of
        :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` containing
        the patches to follow for image correlation.
    
    .. versionadded:: 2.0.0 *log_queue*, *log_level* and *max_freq* arguments
    """

    self._patch_size: Optional[CameraScaleSetting] = None

    super().__init__(camera, log_queue, log_level, max_freq)

    # Setting the patches
    self._spots = patches

  def finish(self) -> None:
    """Method called when the user tries to close the configuration window.

    Check that patches were selected on the image. If not, warns the user and
    prevents him from exiting except with CTRL+C. If not already done by the
    user, also saves the initial length between the patches.
    
    .. versionadded:: 2.0.0
    """

    if self._spots.empty():
      self.log(logging.WARNING, "No patches selected ! Not exiting the "
                                "configuration window")
      showerror('Error !',
                message="Please select patches before exiting the config "
                        "window !\nOr hit CTRL+C to exit Crappy")
      return

    self._spots.save_length()
    self.log(logging.INFO,
             f"Successfully saved L0 ! L0 x : {self._spots.x_l0}, "
             f"L0 y : {self._spots.y_l0}")

    super().stop()

  def _add_settings(self) -> None:
    """Same as in the parent class except it also adds a Path size setting to
    the list of possible settings."""

    self._patch_size = CameraScaleSetting("Patch size (px)", 2, 1024,
                                          default=128)
    self._add_slider_setting(self._patch_size)

    super()._add_settings()

  def _update_settings(self) -> None:
    """Same as in the parent class except it also updates the Path size setting
    in addition to all the other settings."""

    if self._patch_size.value != self._patch_size.tk_var.get():
      self._patch_size.value = self._patch_size.tk_var.get()
    self._patch_size.tk_var.set(self._patch_size.value)

    super()._update_settings()

  def _set_bindings(self) -> None:
    """Binds the left mouse button click to drawing the patches on which to
    perform the image correlation."""

    super()._set_bindings()

    self._img_canvas.bind('<ButtonPress-1>', self._start_box)
    self._img_canvas.bind('<B1-Motion>', self._extend_box)
    self._img_canvas.bind('<ButtonRelease-1>', self._stop_box)

  def _extend_box(self, event: tk.Event) -> None:
    """When the user drags the selection box, updating the four patches being
    drawn."""

    super()._extend_box(event)

    if not self._select_box.no_points() and self._patch_size is not None:
      min_x, max_x, min_y, max_y = self._select_box.sorted()
      size = self._patch_size.value
      if max_x - min_x >= 3 * size and max_y - min_y >= 3 * size:
        self._spots.spot_1 = Box(min_x, min_x + size,
                                 (min_y + max_y - size) // 2,
                                 (min_y + max_y + size) // 2)
        self._spots.spot_2 = Box(max_x - size, max_x,
                                 (min_y + max_y - size) // 2,
                                 (min_y + max_y + size) // 2)
        self._spots.spot_3 = Box((min_x + max_x - size) // 2,
                                 (min_x + max_x + size) // 2,
                                 min_y, min_y + size)
        self._spots.spot_4 = Box((min_x + max_x - size) // 2,
                                 (min_x + max_x + size) // 2,
                                 max_y - size, max_y)

  def _stop_box(self, _: tk.Event) -> None:
    """Simply resets the selection box."""

    # This box is not needed anymore
    self._select_box.reset()

  def _on_img_resize(self, _: Optional[tk.Event] = None) -> None:
    """Same as in the parent class except it also draws the patches on top of
    the displayed image."""

    self.log(logging.DEBUG, "The image canvas was resized")

    self._draw_spots()
    self._resize_img()
    self._display_img()
    self.update()

  def _update_img(self) -> None:
    """Same as in the parent class except it also draws the patches on top of
    the displayed image."""

    self.log(logging.DEBUG, "Updating the image")

    ret = self._camera.get_image()

    # If no frame could be grabbed from the camera
    if ret is None:
      # If it's the first call, generate error image to initialize the window
      if not self._n_loops:
        self.log(logging.WARNING, "Could not get an image from the camera, "
                                  "displaying an error image instead")
        ret = None, np.array(Image.open(BytesIO(resource_string(
          'crappy', 'tool/data/no_image.png'))))
      # Otherwise, just pass
      else:
        self.log(logging.DEBUG, "No image returned by the camera")
        self.update()
        sleep(0.001)
        return

    self._n_loops += 1
    _, img = ret

    if img.dtype != self.dtype:
      self.dtype = img.dtype
    if self.shape != img.shape:
      self.shape = img.shape

    self._cast_img(img)
    self._draw_spots()
    self._resize_img()

    self._calc_hist()
    self._resize_hist()

    self._display_img()
    self._display_hist()

    self._update_pixel_value()

    self.update()

  def _handle_box_outside_img(self, box: Box) -> None:
    """If a patch is outside the image, warning the user and resetting the
    patches."""

    self.log(logging.WARNING, f"The patch {box} is outside the image, "
                              f"resetting the patches")
    self._spots.reset()



# coding: utf-8

from __future__ import annotations
from typing import Optional, List


class I2CMessage:
  """Class that mimics the :class:`smbus2.i2c_msg` class.

  It is used for communication with the
  :class:`~crappy.tool.ft232h.FT232HServer`, only by the
  :class:`~crappy.inout.MPRLS` InOut.
  
  .. versionadded:: 1.5.10
  .. versionchanged:: 2.0.0 renamed from i2c_msg_ft232h to I2CMessage
  """

  def __init__(self,
               type_: str,
               address: int,
               length: Optional[int] = None,
               buf: Optional[list] = None) -> None:
    """Simply sets the attributes of the class, that characterizes the I2C
    message.

    Args:
      type_: Either a read (`'r'`) or a write (`'w'`) message
      address: The address of the I2C slave, as an :obj:`int`.
      length: For a read message, the number of bytes to read as an :obj:`int`.
      buf: For a write message, the :obj:`list` of :obj:`bytes` to be written.
    """

    if type_ not in ['r', 'w']:
      raise ValueError("type_ should be either 'r' or 'w' !")

    self.type = type_
    self.addr = address
    self.len = length if length else 0
    self.buf = buf if buf else []

  @classmethod
  def read(cls, address: int, length: int) -> I2CMessage:
    """Instantiates an :class:`I2CMessage` object for reading bytes.

    Args:
      address: The address of the I2C slave, as an :obj:`int`.
      length: The number of bytes to read, as an :obj:`int`.
    """

    return cls(type_='r', address=address, length=length)

  @classmethod
  def write(cls, address: int, buf: list) -> I2CMessage:
    """Instantiates an :class:`I2CMessage` object for writing bytes.

    Args:
      address: The address of the I2C slave, as an :obj:`int`.
      buf: The :obj:`list` of :obj:`bytes` to be written.
    """

    return cls(type_='w', address=address, buf=buf)

  @property
  def addr(self) -> int:
    """The address of the I2C slave."""

    return self._addr

  @addr.setter
  def addr(self, addr_: int) -> None:
    if not isinstance(addr_, int) or not 0 <= addr_ <= 127:
      raise ValueError("addr should be an integer between 0 and 127 !")
    self._addr = addr_

  @property
  def buf(self) -> List[int]:
    """The list of bytes to be written, or the list of bytes read."""

    return self._buf

  @buf.setter
  def buf(self, buf_: List[int]) -> None:
    if self.type == 'w' and not buf_:
      raise ValueError("buf can't be empty for a write operation !")
    self._buf = buf_

  @property
  def len(self) -> int:
    """The number of bytes to read."""

    return self._len

  @len.setter
  def len(self, len_: int) -> None:
    if self.type == 'r' and (not isinstance(len_, int) or not len_ > 0):
      raise ValueError("len cannot be zero for a read operation !")
    self._len = len_

  def __iter__(self) -> object:
    self._n = 0
    return self

  def __next__(self) -> int:
    try:
      item = self._buf[self._n]
    except IndexError:
      raise StopIteration
    self._n += 1
    return item



# coding: utf-8

from csv import DictWriter
import numpy as np
from typing import Optional, Union
from pathlib import Path
import logging
import logging.handlers

from .camera_process import CameraProcess
from ..._global import OptionalModule

try:
  import SimpleITK as Sitk
except (ModuleNotFoundError, ImportError):
  Sitk = OptionalModule("SimpleITK")

try:
  import PIL
  from PIL.ExifTags import TAGS
  TAGS_INV = {val: key for key, val in TAGS.items()}
except (ModuleNotFoundError, ImportError):
  PIL = OptionalModule("Pillow")
  TAGS = TAGS_INV = OptionalModule("Pillow")

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class ImageSaver(CameraProcess):
  """This :class:`~crappy.blocks.camera_processes.CameraProcess` can record
  images acquired by a :class:`~crappy.blocks.Camera` Block to the desired
  location and in the desired format.

  Various backends can be used for recording the images, some may be faster or
  slower depending on the machine. It is possible to only save one out of a
  given number of images, if not all frames are needed.

  .. versionadded:: 2.0.0
  """

  def __init__(self,
               img_extension: str = "tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      img_extension: The file extension for the recorded images, as a
        :obj:`str` and without the dot. Common file extensions include `tiff`,
        `png`, `jpg`, etc.
      save_folder: Path to the folder where to save the images. Can be an
        absolute or a relative path. The folder does not need to already exit,
        in which case it is created.
      save_period: Only one out of that number images at most will be saved.
        Allows to have a known periodicity in case the framerate is too high
        to record all the images. Or simply to reduce the number of saved
        images if saving them all is not needed.
      save_backend: The backend to use for saving the images. Should be one of:
        ::

          'sitk', 'pil', 'cv2', 'npy'

        They correspond to the modules :mod:`SimpleITK`, :mod:`PIL` (Pillow
        Fork), :mod:`cv2` (OpenCV), and :mod:`numpy`. Depending on the machine,
        some may be faster or slower. The ``img_extension`` is ignored for the
        backend ``'npy'``, that saves the images as raw numpy arrays.
    """

    super().__init__()

    # Trying the different possible backends and checking if the given one
    # is correct
    if save_backend is None:
      if not isinstance(Sitk, OptionalModule):
        self._save_backend = 'sitk'
      elif not isinstance(PIL, OptionalModule):
        self._save_backend = 'pil'
      elif not isinstance(cv2, OptionalModule):
        self._save_backend = 'cv2'
      else:
        self._save_backend = 'npy'
    elif save_backend in ('sitk', 'pil', 'cv2', 'npy'):
      self._save_backend = save_backend
    else:
      raise ValueError("The save_backend argument should be either 'sitk', "
                       "'pil', 'cv2' or 'npy' !")

    # In case the images are saved as arrays, don't include extension
    self._img_extension = img_extension if self._save_backend != 'npy' else ''

    # Setting a default save folder if not given
    if save_folder is None:
      self._save_folder = Path.cwd() / 'Crappy_images'
    else:
      self._save_folder = Path(save_folder)

    self._save_period = int(save_period)

    self._csv_created = False
    self._csv_path = None
    self._metadata_name = 'metadata.csv'

  def init(self) -> None:
    """Creates the folder for saving the images.

    If a folder is already present at the indicated path and contains images,
    saving to a new folder with the same name but ending with a suffix.
    """

    # If the save folder already exists, checking if it contains images by
    # checking if a metadata file is present
    if self._save_folder.exists():
      content = (path.name for path in self._save_folder.iterdir())
      # If it contains images, saving to a different folder
      if self._metadata_name in content:
        self.log(logging.WARNING, f"The folder {self._save_folder} already "
                                  f"seems to contain images from Crappy !")
        parent, name = self._save_folder.parent, self._save_folder.name
        i = 1
        # Adding an integer at the end of the folder name to differentiate it
        while (parent / f'{name}_{i:05d}').exists():
          i += 1
        self._save_folder = parent / f'{name}_{i:05d}'
        self.log(logging.WARNING, f"Saving the images at {self._save_folder} "
                                  f"instead !")

      else:
        self.log(logging.DEBUG,
                 f"The folder {self._save_folder} for recording images exists"
                 f" but does not contain images yet.")

    # Creating the folder for recording images
    if not self._save_folder.exists():
      self.log(logging.INFO, f"Creating the folder for saving images at: "
                             f"{self._save_folder}")
      Path.mkdir(self._save_folder, exist_ok=True, parents=True)

  def _get_data(self) -> bool:
    """Method similar to the one of the parent class, except it also ensures
    that at most only one out of ``save_period`` images is being saved.

    Returns:
      :obj:`True` in case a frame was acquired and needs to be handled, or
      :obj:`False` if no frame was grabbed and nothing should be done.
    """

    # Acquiring the Lock to avoid conflicts with other CameraProcesses
    with self._lock:

      # In case there's no frame grabbed yet
      if 'ImageUniqueID' not in self._data_dict:
        return False

      # In case the frame in buffer was already handled during a previous loop,
      if self._data_dict['ImageUniqueID'] == self.metadata['ImageUniqueID']:
        return False

     # In case it's too early to save the new frame
      if self.metadata['ImageUniqueID'] is not None and \
          self._data_dict['ImageUniqueID'] - self.metadata['ImageUniqueID'] \
          < self._save_period:
        return False

      # Copying the metadata
      self.metadata = self._data_dict.copy()

      self.log(logging.DEBUG, f"Got new image to process with id "
                              f"{self.metadata['ImageUniqueID']}")

      # Copying the frame
      np.copyto(self.img,
                np.frombuffer(self._img_array.get_obj(),
                              dtype=self._dtype).reshape(self._shape))

    return True

  def loop(self) -> None:
    """This method grabs the latest frame, writes its metadata to a `.csv` file
    and saves the image at the chosen location using the chosen backend.

    On the first frame, the metadata file is created and its header is
    populated using the metadata of the frame.
    """

    # Creating the .csv containing the metadata on the first received frame
    if not self._csv_created:
      self._csv_path = (self._save_folder / self._metadata_name)

      self.log(logging.INFO, f"Creating file for saving the metadata: "
                             f"{self._csv_path}")

      # Also writing the header of the .csv file when creating it
      with open(self._csv_path, 'w') as csvfile:
        writer = DictWriter(csvfile, fieldnames=self.metadata.keys())
        writer.writeheader()

      self._csv_created = True

    # Saving the received metadata to the .csv file
    self.log(logging.DEBUG, f"Saving metadata: {self.metadata}")
    with open(self._csv_path, 'a') as csvfile:
      writer = DictWriter(csvfile, fieldnames=self.metadata.keys())
      writer.writerow({**self.metadata, 't(s)': self.metadata['t(s)']})

    # Only include the extension for the image file if applicable
    if self._img_extension:
      path = str(self._save_folder / f"{self.metadata['ImageUniqueID']}_"
                                     f"{self.metadata['t(s)']:.3f}."
                                     f"{self._img_extension}")
    else:
      path = str(self._save_folder / f"{self.metadata['ImageUniqueID']}_"
                                     f"{self.metadata['t(s)']:.3f}")

    # Saving the image at the destination path using the chosen backend
    self.log(logging.DEBUG, "Saving image")
    if self._save_backend == 'sitk':
      Sitk.WriteImage(Sitk.GetImageFromArray(self.img), path)

    elif self._save_backend == 'cv2':
      cv2.imwrite(path, self.img)

    elif self._save_backend == 'pil':
      PIL.Image.fromarray(self.img).save(
        path, exif={TAGS_INV[key]: val for key, val in self.metadata.items()
                    if key in TAGS_INV})

    elif self._save_backend == 'npy':
      np.save(path, self.img)



# coding: utf-8

from typing import Dict, Type

from .ads1115 import ADS1115
from .agilent_34420A import Agilent34420a
from .comedi import Comedi
from .daqmx import DAQmx
from .fake_inout import FakeInOut
from .gpio_pwm import GPIOPWM
from .gpio_switch import GPIOSwitch
from .kollmorgen_akd_pdmm import KollmorgenAKDPDMM
from .labjack_t7 import LabjackT7
from .labjack_t7_streamer import T7Streamer
from .labjack_ue9 import LabjackUE9
from .mcp9600 import MCP9600
from .mprls import MPRLS
from .nau7802 import NAU7802
from .ni_daqmx import NIDAQmx
from .opsens_handysens import HandySens
from .pijuice_hat import PiJuice
from .sim868 import Sim868
from .spectrum_m2i4711 import SpectrumM2I4711
from .waveshare_ad_da import WaveshareADDA
from .waveshare_high_precision import WaveshareHighPrecision

from .ft232h import ADS1115FT232H
from .ft232h import GPIOSwitchFT232H
from .ft232h import MCP9600FT232H
from .ft232h import MPRLSFT232H
from .ft232h import NAU7802FT232H
from .ft232h import WaveshareADDAFT232H

from .meta_inout import InOut, MetaIO

# All the inout objects
from ._deprecated import deprecated_inouts
inout_dict: Dict[str, Type[InOut]] = MetaIO.classes



# coding: utf-8

from time import time
from typing import Dict, List, Any, Optional, Iterable
from dataclasses import dataclass, fields
import logging

from .meta_block import Block
from ..actuator import actuator_dict, Actuator, deprecated_actuators
from ..tool.ft232h import USBServer


@dataclass
class ActuatorInstance:
  """This class holds all the information that can be associated to an
  Actuator."""

  actuator: Actuator
  speed: Optional[float] = None
  position_label: Optional[str] = None
  speed_label: Optional[str] = None
  mode: str = 'speed'
  cmd_label: str = 'cmd'
  speed_cmd_label: Optional[str] = None


class Machine(Block):
  """This Block is meant to drive one or several
  :class:`~crappy.actuator.Actuator`. It can set speed or position commands on
  hardware actuators.

  The possibility to drive several Actuators from a unique Block is given so
  that they can be driven in a synchronized way. If synchronization is not
  needed, it is preferable to drive the Actuators from separate Machine Blocks.

  This Block takes the speed or position commands for the Actuators  as inputs,
  and can optionally read and output the current speed and/or positions of the
  Actuators. The speed and position commands are set respectively by calling
  the :meth:`~crappy.actuator.Actuator.set_position` and
  :meth:`~crappy.actuator.Actuator.set_speed` methods of the Actuators, and the
  current speed and position values are acquired by calling the
  :meth:`~crappy.actuator.Actuator.get_position` and
  :meth:`~crappy.actuator.Actuator.get_speed` methods of the Actuators.

  It is possible to tune for each Actuator the label over which it receives its
  commands, and optionally the labels over which it sends its current speed
  and/or position. The driving mode (`'speed'` or `'position'`) can also be set
  independently for each Actuator.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               actuators: Iterable[Dict[str, Any]],
               common: Optional[Dict[str, Any]] = None,
               time_label: str = 't(s)',
               ft232h_ser_num: Optional[str] = None,
               spam: bool = False,
               freq: Optional[float] = 200,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      actuators: An iterable (like a :obj:`list` or a :obj:`tuple`) of all the
        :class:`~crappy.actuator.Actuator` this Block needs to drive. It 
        contains one :obj:`dict` for every Actuator, with mandatory and 
        optional keys. The keys providing information on how to drive the 
        Actuator are listed below. Any other unrecognized key will be passed to 
        the Actuator as argument when instantiating it.
      common: The keys of this :obj:`dict` will be common to all the Actuators.
        If one key conflicts with an existing key for an Actuator, the common 
        one will prevail.
      time_label: If reading speed or position from one or more Actuators, the
        time information will be carried by this label.
      spam: If :obj:`True`, a command is sent to the Actuators at each loop of
        the Block, else it is sent every time a new command is received.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.

    Note:
      - ``actuators`` keys:

        - ``type``: The name of the :class:`~crappy.actuator.Actuator` class to 
          instantiate. This key is mandatory.
        - ``cmd_label``: The label carrying the command for driving the
          Actuator. It defaults to `'cmd'`.
        - ``mode``: Can be either `'speed'` or `'position'`. Either
          :meth:`~crappy.actuator.Actuator.set_speed` or
          :meth:`~crappy.actuator.Actuator.set_position` is called to drive the
          Actuator, depending on the selected mode. When driven in `'position'` 
          mode, the speed of the actuator can also be adjusted, see the 
          ``speed`` and ``speed_cmd_label`` keys. The default mode is 
          `'speed'`.
        - ``speed``: If mode is `'position'`, the speed at which the Actuator
          should move. This speed is passed as second argument to the
          :meth:`~crappy.actuator.Actuator.set_position` method of the
          Actuator. If the ``speed_cmd_label`` key is not specified, this speed
          will remain the same for the entire test. This key is not mandatory.
        - ``position_label``: If given, the Block will return the value of
          :meth:`~crappy.actuator.Actuator.get_position` under this label. This
          key is not mandatory.
        - ``speed_label``: If given, the Block will return the value of
          :meth:`~crappy.actuator.Actuator.get_speed` under this label. This
          key is not mandatory.
        - ``speed_cmd_label``: The label carrying the speed to set when driving
          in `'position'` mode. Each time a value is received, the stored speed
          value is updated. It will also overwrite the ``speed`` key if given.

    .. versionadded:: 1.5.10 *verbose* argument
    .. versionadded 2.0.0 *debug* and *ft232h_ser_num* arguments
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    self._actuators: List[ActuatorInstance] = list()
    self._ft232h_args = None

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    self._time_label = time_label
    self._spam = spam

    # No extra information to add to the main dicts
    if common is None:
      common = dict()

    # Updating the settings with the common information
    for actuator in actuators:
      actuator.update(common)

    # Making sure all the dicts contain the 'type' key
    if not all('type' in dic for dic in actuators):
      raise ValueError("The 'type' key must be provided for all the "
                       "actuators !")

    # The names of the possible settings, to avoid typos and reduce verbosity
    actuator_settings = [field.name for field in fields(ActuatorInstance)
                         if field.type is not Actuator]

    # The list of all the Actuator types to instantiate
    self._types = [actuator['type'] for actuator in actuators]

    # Checking for deprecated names
    deprecated = [type_ for type_ in self._types
                  if type_ in deprecated_actuators]
    for type_ in deprecated:
      raise NotImplementedError(
          f"The {type_} Actuator was deprecated in version 2.0.0, and renamed "
          f"to {deprecated_actuators[type_]} ! Please update your code "
          f"accordingly and check the documentation for more information")

    # Checking that all the given actuators are valid
    if not all(type_ in actuator_dict for type_ in self._types):
      unknown = ', '.join(tuple(type_ for type_ in self._types if type_
                          not in actuator_dict))
      possible = ', '.join(sorted(actuator_dict.keys()))
      raise ValueError(f"Unknown actuator type(s) : {unknown} ! "
                       f"The possible types are : {possible}")

    # The settings that won't be passed to the Actuator objects
    self._settings = [{key: value for key, value in actuator.items()
                       if key in actuator_settings}
                      for actuator in actuators]

    # The settings that will be passed as kwargs to the Actuator objects
    self._actuators_kw = [{key: value for key, value in actuator.items()
                           if key not in ('type', *actuator_settings)}
                          for actuator in actuators]

    # Checking whether the Actuators communicate through an FT232H
    if any(actuator_dict[type_].ft232h for type_ in self._types):
      self._ft232h_args = USBServer.register(ft232h_ser_num)

  def prepare(self) -> None:
    """Checks the validity of the linking and initializes all the Actuator
    objects to drive.

    This method calls the :meth:`~crappy.actuator.Actuator.open` method of each
    Actuator.
    """

    # Instantiating the actuators and storing them
    self._actuators = [ActuatorInstance(
      actuator=actuator_dict[type_](**actuator_kw)
      if not actuator_dict[type_].ft232h else
      actuator_dict[type_](**actuator_kw, _ft232h_args=self._ft232h_args),
      **setting)
      for type_, setting, actuator_kw in zip(self._types,
                                             self._settings,
                                             self._actuators_kw)]

    # Checking the consistency of the linking
    if not self.inputs and not self.outputs:
      raise IOError("The Machine block isn't linked to any other block !")

    # Opening each actuator
    for actuator in self._actuators:
      self.log(logging.INFO, f"Opening the {type(actuator.actuator).__name__}"
                             f"Actuator")
      actuator.actuator.open()
      self.log(logging.INFO, f"Opened the {type(actuator.actuator).__name__}"
                             f"Actuator")

  def loop(self) -> None:
    """Sets the received position and speed commands, and reads the current 
    speed and position from the :class:`~crappy.actuator.Actuator`.
    
    For each Actuator, a command is set **only** if a new one was received or 
    if the ``spam`` argument is :obj:`True`. It is set using either 
    :meth:`~crappy.actuator.Actuator.set_position` or
    :meth:`~crappy.actuator.Actuator.set_speed` depending on the selected
    driving mode.
    
    For each Actuator, a speed and/or position value is read **only** if the 
    ``speed_label`` and/or the ``position_label`` was set. If so, these values
    are read at each loop and sent to downstream Blocks over the given labels.
    This is independent of the chosen driving mode. The
    :meth:`~crappy.actuator.Actuator.get_position` and
    :meth:`~crappy.actuator.Actuator.get_speed` are called for acquiring the
    position and speed values respectively.
    """

    # Receiving the latest command
    recv = self.recv_last_data(fill_missing=self._spam)

    # Iterating over the actuators for setting the commands
    if recv:
      for actuator in self._actuators:
        # Setting the speed attribute if it was received
        if actuator.speed_cmd_label is not None and \
            actuator.speed_cmd_label in recv:
          self.log(logging.DEBUG,
                   f"Updating the speed of the "
                   f"{type(actuator.actuator).__name__} Actuator from "
                   f"{actuator.speed} to {recv[actuator.speed_cmd_label]}")
          actuator.speed = recv[actuator.speed_cmd_label]

        # Setting only the commands that were received
        if actuator.cmd_label in recv:
          # Setting the speed command
          if actuator.mode == 'speed':
            self.log(logging.DEBUG,
                     f"Setting speed of the {type(actuator.actuator).__name__}"
                     f" Actuator to {recv[actuator.cmd_label]}")
            actuator.actuator.set_speed(recv[actuator.cmd_label])
          # Setting the position command
          else:
            actuator.actuator.set_position(recv[actuator.cmd_label],
                                           actuator.speed)
            self.log(
              logging.DEBUG,
              f"Setting position of the {type(actuator.actuator).__name__} "
              f"Actuator to {recv[actuator.cmd_label]} with speed "
              f"{actuator.speed}")

    to_send = {}

    # Iterating over the actuators to get the speeds and the positions
    for actuator in self._actuators:
      if actuator.position_label is not None:
        position = actuator.actuator.get_position()
        if position is not None:
          to_send[actuator.position_label] = position
      if actuator.speed_label is not None:
        speed = actuator.actuator.get_speed()
        if speed is not None:
          to_send[actuator.speed_label] = speed

    # Sending the speed and position values if any
    if to_send:
      to_send[self._time_label] = time() - self.t0
      self.send(to_send)

  def finish(self) -> None:
    """Stops and closes all the Actuators to drive.

    This method calls the :meth:`~crappy.actuator.Actuator.stop` and
    :meth:`~crappy.actuator.Actuator.close` method of each Actuator.
    """

    for actuator in self._actuators:
      self.log(logging.INFO, f"Stopping the {type(actuator.actuator).__name__}"
                             f"Actuator")
      actuator.actuator.stop()
    for actuator in self._actuators:
      self.log(logging.INFO, f"Closing the {type(actuator.actuator).__name__}"
                             f"Actuator")
      actuator.actuator.close()
      self.log(logging.INFO, f"Closed the {type(actuator.actuator).__name__}"
                             f"Actuator")



# coding: utf-8

from time import time
from typing import List
import logging
from  warnings import warn

from ..meta_inout import InOut
from ...tool.ft232h import FT232HServer as FT232H, USBArgsType

Mcp9600_registers = {'Hot Junction Temperature': 0x00,
                     'Junction Temperature Delta': 0x01,
                     'Cold Junction Temperature': 0x02,
                     'Raw Data ADC': 0x03,
                     'Status': 0x04,
                     'Thermocouple Sensor Configuration': 0x05,
                     'Device Configuration': 0x06,
                     'Alert 1 Configuration': 0x08,
                     'Alert 2 Configuration': 0x09,
                     'Alert 3 Configuration': 0x0A,
                     'Alert 4 Configuration': 0x0B,
                     'Alert 1 Hysteresis': 0x0C,
                     'Alert 2 Hysteresis': 0x0D,
                     'Alert 3 Hysteresis': 0x0E,
                     'Alert 4 Hysteresis': 0x0F,
                     'Alert 1 Limit': 0x10,
                     'Alert 2 Limit': 0x11,
                     'Alert 3 Limit': 0x12,
                     'Alert 4 Limit': 0x13,
                     'Device ID/REV': 0x20}

Mcp9600_filter_coefficients = {0: 0b000,
                               1: 0b001,
                               2: 0b010,
                               3: 0b011,
                               4: 0b100,
                               5: 0b101,
                               6: 0b110,
                               7: 0b111}

Mcp9600_thermocouple_types = {'J': 0b0000000,
                              'K': 0b0010000,
                              'T': 0b0100000,
                              'N': 0b0110000,
                              'S': 0b1000000,
                              'E': 0b1010000,
                              'B': 0b1100000,
                              'R': 0b1110000}

Mcp9600_adc_resolutions = {18: 0b0000000,
                           16: 0b0100000,
                           14: 0b1000000,
                           12: 0b1100000}

Mcp9600_sensor_resolutions = {0.0625: 0b00000000,
                              0.25: 0b10000000}

Mcp9600_modes = ['Hot Junction Temperature',
                 'Junction Temperature Delta',
                 'Cold Junction Temperature',
                 'Raw Data ADC']


class MCP9600FT232H(InOut):
  """This class can read temperature values from an MCP9600 thermocouple
  reader through an FT232H.

  It is similar to the :class:`~crappy.inout.MCP9600` class, except this class
  is specific for use with an :class:`~crappy.tool.ft232h.FT232H` USB to I2C
  converter.

  It communicates over the I2C protocol. The output is in `°C`, except for one
  operating mode that returns Volts. Several parameters can be tuned, like the
  thermocouple type, the reading resolution or the filter coefficient. Note
  that the MCP9600 can only achieve a data rate of a few Hz.

  .. versionadded:: 2.0.0
  """

  ft232h = True

  def __init__(self,
               thermocouple_type: str,
               device_address: int = 0x67,
               adc_resolution: int = 18,
               sensor_resolution: float = 0.0625,
               filter_coefficient: int = 0,
               mode: str = 'Hot Junction Temperature',
               _ft232h_args: USBArgsType = tuple()) -> None:
    """Checks the validity of the arguments.

    Args:
      thermocouple_type: The type of thermocouple connected to the MCP9600. The
        possible types are:
        ::

          'J', 'K', 'T', 'N', 'S', 'E', 'B', 'R'

      device_address: The I2C address of the MCP9600. The default address is
        `0x67`, but it is possible to change this setting using a specific
        setup involving the `ADDR` pin.
      adc_resolution: The number of bits the ADC output is encoded on. The
        greater the resolution, the lower the sample rate. The available
        resolutions are:
        ::

          12, 14, 16, 18

      sensor_resolution: The temperature measurement resolution in `°C`. It
        should be either `0.0625` or `0.25`. Setting the resolution to `0.25`
        will increase the sample rate, but the output temperature will be
        encoded on two bits less.
      filter_coefficient: The MCP9600 features an integrated filter (see its
        documentation for the exact filter formula). When set to `0`, the
        filter is inactive. It is maximal when set to `7`. When active, the
        filter will prohibit fast temperature changes, thus limiting noise and
        smoothening the signal.
      mode: Four different values can be accessed when measuring a temperature:
        the temperature of the thermocouple (hot junction temperature), the
        temperature of the MCP9600 board (cold junction temperature), the
        temperature calculated from the ADC data and thermocouple type but not
        yet cold junction-compensated (junction temperature delta), and the raw
        ADC measurement of the voltage difference in the thermocouple (raw data
        ADC, in Volts). The available modes are thus:
        ::

          'Hot Junction Temperature',
          'Junction Temperature Delta',
          'Cold Junction Temperature',
          'Raw Data ADC'

      _ft232h_args: This argument is meant for internal use only and should not
        be provided by the user. It contains the information necessary for
        setting up the FT232H.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._bus = None

    super().__init__()

    (block_index, block_lock, command_file, answer_file, shared_lock,
     current_block) = _ft232h_args

    self._bus = FT232H(mode='I2C',
                       block_index=block_index,
                       current_block=current_block,
                       command_file=command_file,
                       answer_file=answer_file,
                       block_lock=block_lock,
                       shared_lock=shared_lock,
                       i2c_speed=20E3)

    self._device_address = device_address

    if sensor_resolution not in Mcp9600_sensor_resolutions:
      raise ValueError("sensor_resolution should be in {}".format(list(
        Mcp9600_sensor_resolutions.keys())))
    else:
      self._sensor_resolution = sensor_resolution

    if adc_resolution not in Mcp9600_adc_resolutions:
      raise ValueError("adc_resolution should be in {}".format(list(
        Mcp9600_adc_resolutions.keys())))
    else:
      self._adc_resolution = adc_resolution

    if thermocouple_type not in Mcp9600_thermocouple_types:
      raise ValueError("thermocouple_type should be in {}".format(list(
        Mcp9600_thermocouple_types.keys())))
    else:
      self._thermocouple_type = thermocouple_type

    if filter_coefficient not in Mcp9600_filter_coefficients:
      raise ValueError("filter_coefficient should be an integer between "
                       "0 and 7")
    else:
      self._filter_coefficient = filter_coefficient

    if mode not in Mcp9600_modes:
      raise ValueError("mode should be in {}".format(Mcp9600_modes))
    else:
      self._mode = mode

  def open(self) -> None:
    """Initializes the I2C communication and the device."""

    if not self._is_connected():
      raise IOError("The MCP9600 is not connected")
    self.log(logging.INFO, "Setting up the MCP9600")

    # Setting the sensor according to the user parameters
    config_sensor = Mcp9600_filter_coefficients[self._filter_coefficient]
    config_sensor |= Mcp9600_thermocouple_types[self._thermocouple_type]

    self.log(logging.DEBUG,
             f"Writing {config_sensor} to the register "
             f"{Mcp9600_registers['Thermocouple Sensor Configuration']} "
             f"at address {self._device_address}")
    self._bus.write_i2c_block_data(self._device_address, Mcp9600_registers[
        'Thermocouple Sensor Configuration'], [config_sensor])

    # Setting the device according to the user parameters
    config_device = Mcp9600_sensor_resolutions[self._sensor_resolution]
    config_device |= Mcp9600_adc_resolutions[self._adc_resolution]
    config_device |= 0b00  # Normal operating mode
    self.log(logging.DEBUG,
             f"Writing {config_device} to the register "
             f"{Mcp9600_registers['Device Configuration']} "
             f"at address {self._device_address}")
    self._bus.write_i2c_block_data(self._device_address,
                                   Mcp9600_registers['Device Configuration'],
                                   [config_device])

  def get_data(self) -> List[float]:
    """Reads the registers containing the conversion result.

    The output is in `°C` for all modes except the raw data ADC one, which
    outputs Volts.

    Returns:
      A :obj:`list`: containing the timestamp and the output value.
    """

    # Starting a conversion
    value = self._bus.read_i2c_block_data(self._device_address,
                                          Mcp9600_registers['Status'], 1)[0]
    self.log(logging.DEBUG, f"Read {value} from register "
                            f"{Mcp9600_registers['Status']} at address "
                            f"{self._device_address}")
    self.log(logging.DEBUG,
             f"Writing {value & 0xBF} to the register "
             f"{Mcp9600_registers['Status']} at address "
             f"{self._device_address}")
    self._bus.write_i2c_block_data(self._device_address,
                                   Mcp9600_registers['Status'],
                                   [value & 0xBF])

    # Waiting for the conversion to complete
    t0 = time()
    while not self._data_available():
      if time() - t0 > 1:
        raise TimeoutError('Waited too long for data to be ready !')

    out = [time()]

    # The number of output bits varies according to the selected mode
    # Temperature-modes outputs are 2 bytes long
    if self._mode in ['Hot Junction Temperature',
                      'Junction Temperature Delta',
                      'Cold Junction Temperature']:
      block = self._bus.read_i2c_block_data(self._device_address,
                                            Mcp9600_registers[self._mode], 2)
      self.log(logging.DEBUG, f"Read {block} from register "
                              f"{Mcp9600_registers[self._mode]} at address "
                              f"{self._device_address}")
      value_raw = ((block[0] << 8) | block[1])

      # Converting the raw output value into °C
      value = value_raw / 16
      if block[0] >> 7:
        value -= 4096
      out.append(value)

    # Raw ADC output is 3 bytes long
    else:
      block = self._bus.read_i2c_block_data(self._device_address,
                                            Mcp9600_registers[self._mode], 3)
      self.log(logging.DEBUG, f"Read {block} from register "
                              f"{Mcp9600_registers[self._mode]} at address "
                              f"{self._device_address}")
      value_raw = (block[0] << 16) | (block[1] << 8) | block[2]

      # Converting the raw output value into Volts
      # The 2µV resolution is given in the MCP9600 datasheet
      if block[0] >> 1 & 1:
        value_raw -= 2 ** 18

      out.append(value_raw * 2E-6)

    return out

  def close(self) -> None:
    """Switches the MCP9600 to shutdown mode and closes the I2C bus."""

    if self._bus is not None:
      # Switching to shut down mode, keeping configuration
      value = self._bus.read_i2c_block_data(self._device_address,
                                            Mcp9600_registers[
                                             'Device Configuration'], 1)[0]
      self.log(logging.DEBUG, f"Read {value} from register "
                              f"{Mcp9600_registers['Device Configuration']}"
                              f" at address {self._device_address}")
      value &= 0xFD
      value |= 0x01
      self.log(logging.DEBUG,
               f"Writing {value} to the register "
               f"{Mcp9600_registers['Device Configuration']} at address "
               f"{self._device_address}")
      self._bus.write_i2c_block_data(self._device_address,
                                     Mcp9600_registers['Device Configuration'],
                                     [value])
      self.log(logging.INFO, "Closing the I2C connection to the MCP9600")
      self._bus.close()

  def _data_available(self) -> int:
    """Reads the data available bit.

    Returns: `1` if data is available, else `0`
    """

    status = self._bus.read_i2c_block_data(self._device_address,
                                           Mcp9600_registers['Status'], 1)[0]
    self.log(logging.DEBUG,
             f"Read {status} from register {Mcp9600_registers['Status']}"
             f" at address {self._device_address}")

    # The MCP9600 features an over-temperature protection
    if status & 0x10:
      raise ValueError("Too hot for selected thermocouple !")

    return status & 0x40

  def _is_connected(self) -> bool:
    """Tries reading a byte from the device.

    Returns: :obj:`True` if reading was successful, else :obj:`False`
    """

    try:
      self._bus.read_byte(self._device_address)
      return True
    except IOError:
      return False



# coding:utf-8

from .constant import Constant
from .cyclic import Cyclic
from .ramp import Ramp
from .cyclic_ramp import CyclicRamp
from .integrator import Integrator
from .sine import Sine
from .conditional import Conditional
from .custom import Custom

from . import meta_path



# coding: utf-8

import numpy as np
from typing import List, Tuple
from ..._global import OptionalModule
from ..camera_config import SpotsBoxes, Box

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class DICVETool:
  """This class is the core of the :class:`~crappy.blocks.DICVE` Block.

  It tracks patches on images received from a :class:`~crappy.camera.Camera` 
  object, and computes a strain value at each new image.

  It relies on cross-correlation algorithms to calculate the displacement.
  Different algorithms are available depending on the needs. This tool is
  mainly used to perform video-extensometry on speckled surfaces, although it
  could as well be of use for other applications.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from DISVE to DICVETool
  """

  def __init__(self,
               patches: SpotsBoxes,
               method: str = 'Disflow',
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               iterations: int = 1,
               gradient_iterations: int = 10,
               patch_size: int = 8,
               patch_stride: int = 3,
               border: float = 0.2,
               safe: bool = True,
               follow: bool = True) -> None:
    """Sets a few attributes and initializes DISFlow if this method was
    selected.

    Args:
      patches: An instance of the
        :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` class,
        containing the coordinates of the patches to track.
      method: The method to use to calculate the displacement. `Disflow` uses
        opencv's DISOpticalFlow and `Lucas Kanade` uses opencv's
        calcOpticalFlowPyrLK, while all other methods are based on a basic
        cross-correlation in the Fourier domain. `Pixel precision` calculates
        the displacement by getting the position of the maximum of the
        cross-correlation, and has thus a 1-pixel resolution. It is mainly
        meant for debugging. `Parabola` refines the result of
        `Pixel precision` by interpolating the neighborhood of the maximum, and
        has thus a sub-pixel resolution.
      alpha: Weight of the smoothness term in DISFlow, as a :obj:`float`.
      delta: Weight of the color constancy term in DISFlow, as a :obj:`float`.
      gamma: Weight of the gradient constancy term in DISFlow , as a
        :obj:`float`.
      finest_scale: Finest level of the Gaussian pyramid on which the flow
        is computed in DISFlow (`0` means full scale), as an :obj:`int`.
      iterations: Maximum number of gradient descent iterations in the
        patch inverse search stage in DISFlow, as an :obj:`int`.
      gradient_iterations: Maximum number of gradient descent iterations
        in the patch inverse search stage in DISFlow, as an :obj:`int`.
      patch_size: Size of an image patch for matching in DISFlow
        (in pixels).
      patch_stride: Stride between neighbor patches in DISFlow. Must be
        less than patch size.
      border: Crop the patch on each side according to this value before
        calculating the displacements. 0 means no cropping, 1 means the entire
        patch is cropped.
      safe: If :obj:`True`, checks whether the patches aren't exiting the
        image, and raises an error if that's the case.
      follow: It :obj:`True`, the patches will move to follow the displacement
        of the image.
    
    .. versionchanged:: 1.5.9
       renamed *gditerations* argument to *gradient_iterations*
    .. versionadded:: 1.5.9 *method* argument
    .. versionremoved:: 1.5.10 *img0* and *show_image* arguments
    """

    # These attributes are accessed by the parent class
    self.patches = patches
    self._offsets = [(0, 0) for _ in patches]

    # Other attributes to set
    self._method = method
    self._border = border
    self._safe = safe
    self._follow = follow

    # These attributes will be set later
    self._img0 = None
    self._height, self._width = None, None

    # Initialize DISFlow if it is the selected method
    if self._method == 'Disflow':
      self._dis = cv2.DISOpticalFlow_create(cv2.DISOPTICAL_FLOW_PRESET_FAST)
      self._dis.setVariationalRefinementIterations(iterations)
      self._dis.setVariationalRefinementAlpha(alpha)
      self._dis.setVariationalRefinementDelta(delta)
      self._dis.setVariationalRefinementGamma(gamma)
      self._dis.setFinestScale(finest_scale)
      self._dis.setGradientDescentIterations(gradient_iterations)
      self._dis.setPatchSize(patch_size)
      self._dis.setPatchStride(patch_stride)
    else:
      self._dis = None

  def set_img0(self, img0: np.ndarray) -> None:
    """Sets the reference image for the cross-correlation.
    
    .. versionadded:: 1.5.10
    """

    self._img0 = img0
    self._height, self._width, *_ = img0.shape

    # Now that there's an initial image, checking that the patches are valid
    if self._safe:
      self._check_offsets()

  def calculate_displacement(
      self, img: np.ndarray) -> Tuple[List[Tuple[float, float]], float, float,
                                      List[Tuple[float, float]]]:
    """Returns the displacement of every patch, calculated according to the
    chosen method.

    Also updates the patch offsets if required, and updates the window for
    following the patches if any.

    .. versionadded:: 1.5.9
    """

    # Making sure the reference image exists
    if self._img0 is None:
      raise ValueError("The method set_img0 must be called first for setting "
                       "the reference image !")

    # Compute the displacement for each patch
    displacements = []
    for patch, offset in zip(self.patches, self._offsets):

      if patch is None:
        continue

      if self._method == 'Disflow':
        displacements.append(self._calc_disflow(patch, img, offset))

      elif self._method == 'Pixel precision':
        displacements.append(self._calc_pixel_precision(patch, img, offset))

      elif self._method == 'Parabola':
        displacements.append(self._calc_parabola(patch, img, offset))

      elif self._method == 'Lucas Kanade':
        displacements.append(self._calc_lucas_kanade(patch, img, offset))

      else:
        raise ValueError("Wrong method specified !")

    # If required, updates the patch offsets
    if self._follow:
      for disp, (y_offset, x_offset), patch in zip(displacements,
                                                   self._offsets,
                                                   self.patches):

        patch.x_start = round(patch.x_start + disp[0])
        patch.x_end = round(patch.x_end + disp[0])
        patch.y_start = round(patch.y_start + disp[1])
        patch.y_end = round(patch.y_end + disp[1])
        
        patch.x_centroid = (patch.x_end + patch.x_start) / 2
        patch.y_centroid = (patch.y_end + patch.y_start) / 2

        disp[0] += x_offset
        disp[1] += y_offset

        patch.x_disp = disp[0]
        patch.y_disp = disp[1]

      self._offsets = [(round(y_disp), round(x_disp)) for
                       x_disp, y_disp in displacements]

      # Check that the patches are not exiting the image
      if self._safe:
        self._check_offsets()

    else:
      for (x_disp, y_disp), patch in zip(displacements, self.patches):
        patch.x_disp = x_disp
        patch.y_disp = y_disp
    
    max_x = max(self.patches, 
                key=lambda patch_: patch_.x_centroid if patch_ is not None
                else -float('inf'))
    min_x = min(self.patches,
                key=lambda patch_: patch_.x_centroid if patch_ is not None
                else float('inf'))
    max_y = max(self.patches,
                key=lambda patch_: patch_.y_centroid if patch_ is not None
                else -float('inf'))
    min_y = min(self.patches,
                key=lambda patch_: patch_.y_centroid if patch_ is not None
                else float('inf'))

    # If there are multiple spots, the x and y strains can be computed
    if len(self.patches) > 1:
      try:
        exx = ((max_x.x_disp - min_x.x_disp) / self.patches.x_l0) * 100
      except ZeroDivisionError:
        exx = 0
      try:
        eyy = ((max_y.y_disp - min_y.y_disp) / self.patches.y_l0) * 100
      except ZeroDivisionError:
        eyy = 0
      centers = [(patch.y_centroid, patch.x_centroid)
                 for patch in self.patches if patch is not None]
      disps = [(patch.y_disp, patch.x_disp)
               for patch in self.patches if patch is not None]
      return centers, eyy, exx, disps

    # If only one spot was detected, the strain isn't computed
    else:
      x = self.patches[0].x_centroid
      y = self.patches[0].y_centroid
      x_disp = self.patches[0].x_disp
      y_disp = self.patches[0].y_disp
      return [(y, x)], 0, 0, [(y_disp, x_disp)]

  def _calc_disflow(self,
                    patch: Box,
                    img: np.ndarray,
                    offset: Tuple[int, int]) -> List[float]:
    """Returns the displacement between the original and the current image with
    a sub-pixel precision, using DISFlow."""

    disp_img = self._dis.calc(self._get_patch(self._img0, patch, offset),
                              self._get_patch(img, patch), None)
    return np.average(self._trim_patch(disp_img), axis=(0, 1)).tolist()

  def _calc_pixel_precision(self,
                            patch: Box,
                            img: np.ndarray,
                            offset: Tuple[int, int]) -> List[float]:
    """Returns the displacement between the original and the current image with
    a precision limited to 1 pixel."""

    cross_correl, max_width, max_height = self._cross_correlation(
      self._get_patch(self._img0, patch, offset), self._get_patch(img, patch))

    height, width = cross_correl.shape[0], cross_correl.shape[1]
    return [-(max_width - width / 2), -(max_height - height / 2)]

  def _calc_parabola(self,
                     patch: Box,
                     img: np.ndarray,
                     offset: Tuple[int, int]) -> List[float]:
    """Returns the displacement between the original and the current image with
    a sub-pixel precision, using two parabola fits (one in x and one in y)."""

    cross_correl, max_width, max_height = self._cross_correlation(
      self._get_patch(self._img0, patch, offset), self._get_patch(img, patch))

    height, width = cross_correl.shape[0], cross_correl.shape[1]
    y_disp = -(max_height - height / 2)
    x_disp = -(max_width - width / 2)

    x_disp -= self._parabola_fit(cross_correl[max_height,
                                 max_width - 1: max_width + 2])
    y_disp -= self._parabola_fit(cross_correl[max_height - 1: max_height + 2,
                                 max_width])

    return [x_disp, y_disp]

  def _calc_lucas_kanade(self,
                         patch: Box,
                         img: np.ndarray,
                         offset: Tuple[int, int]) -> List[float]:
    """Returns the displacement between the original and the current image with
    a sub-pixel precision, using the Lucas Kanade algorithm."""

    # Getting the center of the patch
    x_top, x_bottom, y_left, y_right = patch.sorted()
    center_y, center_x = (y_right - y_left) // 2, (x_bottom - x_top) // 2

    next_, _, _ = cv2.calcOpticalFlowPyrLK(
      self._get_patch(self._img0, patch, offset), self._get_patch(img, patch),
      np.array([[center_x, center_y]]).astype('float32'), None)
    new_x, new_y = np.squeeze(next_)

    return [new_x - center_x, new_y - center_y]

  @staticmethod
  def _parabola_fit(arr: np.ndarray) -> float:
    """Returns the abscissa of the maximum of a parabola defined by 3
    points in positions x=-1, x=0 and x=1.

    Args:
      arr: This array contains the y values for the 3 points.
    """

    return (arr[0] - arr[2]) / (2 * (arr[0] - 2 * arr[1] + arr[2]))

  @staticmethod
  def _cross_correlation(img0: np.ndarray,
                         img1: np.ndarray) -> Tuple[np.ndarray, int, int]:
    """Performs a cross-correlation operation on two patches in the Fourier
    domain.

    Returns:
      The result of the correlation in the real domain as an image, as well as
      the position of the maximum of this image.
    """

    # Getting the size of the image
    height, width = img0.shape[0], img0.shape[1]

    # Find the closest power of 2 length from the image
    height_log2 = 2 ** int(np.log2(height + 0.5))
    width_log2 = 2 ** int(np.log2(width + 0.5))

    # Cropping the image to a power of 2
    img0 = img0[(height - height_log2) // 2: (height + height_log2) // 2,
                (width - width_log2) // 2: (width + width_log2) // 2]
    img1 = img1[(height - height_log2) // 2: (height + height_log2) // 2,
                (width - width_log2) // 2: (width + width_log2) // 2]

    # Convert to Fourier for fast cross-correlation
    img0_fourier = cv2.dft(img0.astype(np.float32),
                           flags=cv2.DFT_COMPLEX_OUTPUT)
    img1_fourier = cv2.dft(img1.astype(np.float32),
                           flags=cv2.DFT_COMPLEX_OUTPUT)

    # Compute cross-correlation by convolution
    cross_fourier = cv2.mulSpectrums(img0_fourier, img1_fourier,
                                     flags=0, conjB=True)

    # Convert back to physical space
    cross_shifted = cv2.idft(cross_fourier)
    # Un-shift after FFT
    cross_correl = np.fft.ifftshift(cross_shifted[:, :, 0])

    # Find the maximum of the cross-correlation
    max_width, max_height = cv2.minMaxLoc(cross_correl)[-1]
    # Keep the average in case several maxima found
    max_width, max_height = int(np.mean(max_width)), int(np.mean(max_height))

    return cross_correl, max_width, max_height

  @staticmethod
  def _get_patch(img: np.ndarray,
                 patch: Box,
                 offset: Tuple[int, int] = (0, 0)) -> np.ndarray:
    """Returns the part of the image corresponding to the given patch at the
    given offset."""

    y_off, x_off = offset
    x_top, x_bottom, y_left, y_right = patch.sorted()
    return np.array(img[y_left - y_off: y_right - y_off,
                        x_top - x_off: x_bottom - x_off])

  def _trim_patch(self, patch: np.ndarray) -> np.ndarray:
    """Trims the border of a patch according to the value set by the user, and
    returns the sub image corresponding to the trimmed patch."""

    height, width, *_ = patch.shape
    return patch[int(height * self._border / 2):
                 int(height * (1 - self._border / 2)),
                 int(width * self._border / 2):
                 int(width * (1 - self._border / 2))]

  def _check_offsets(self) -> None:
    """Check if the patches are still within the image, and raises an error if
    one of them is out."""

    for patch in self.patches:

      if patch is None:
        continue

      x_top, x_bottom, y_left, y_right = patch.sorted()

      # Checking the left border
      if x_top < 0:
        raise RuntimeError("Region exiting the ROI (left)")

      # Checking the right border
      elif x_bottom > self._width:
        raise RuntimeError("Region exiting the ROI (right)")

      # Checking the top border
      if y_left < 0:
        raise RuntimeError("Region exiting the ROI (top)")

      # Checking the bottom border
      elif y_right > self._height:
        raise RuntimeError("Region exiting the ROI (bottom)")



# coding: utf-8

import tkinter as tk
from tkinter.messagebox import showerror
from typing import Optional
import numpy as np
from io import BytesIO
from pkg_resources import resource_string
from time import sleep
import logging
from multiprocessing.queues import Queue

from .camera_config_boxes import CameraConfigBoxes
from .config_tools import Box
from ...camera.meta_camera import Camera
from ..._global import OptionalModule

try:
  from PIL import Image
except (ModuleNotFoundError, ImportError):
  Image = OptionalModule("pillow")


class DISCorrelConfig(CameraConfigBoxes):
  """Class similar to :class:`~crappy.tool.camera_config.CameraConfig` but also 
  allowing to select the area on which the correlation will be performed.
  
  It relies on the :class:`~crappy.tool.camera_config.config_tools.Box` tool. 
  It is meant to be used for configuring the :class:`~crappy.blocks.DISCorrel` 
  Block.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from DISConfig to DISCorrelConfig
  """

  def __init__(self,
               camera: Camera,
               log_queue: Queue,
               log_level: Optional[int],
               max_freq: Optional[float],
               patch: Box) -> None:
    """Initializes the parent class and sets the correlation Box.

    Args:
      camera: The :class:`~crappy.camera.Camera` object in charge of acquiring 
        the images.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to 
        the main :obj:`~logging.Logger`, only used in Windows.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.
      max_freq: The maximum frequency this window is allowed to loop at. It is
        simply the ``freq`` attribute of the :class:`~crappy.blocks.Camera`
        Block.
      patch: The :class:`~crappy.tool.camera_config.config_tools.Box` container
        that will save the information on the patch where to perform image
        correlation.
        
    .. versionadded:: 2.0.0
       *patch*, *log_queue*, *log_level* and *max_freq* arguments
    """

    self._correl_box = patch
    self._draw_correl_box = True

    super().__init__(camera, log_queue, log_level, max_freq)

  @property
  def box(self) -> Box:
    """Returns the :class:`~crappy.tool.camera_config.config_tools.Box` object
    containing the region of interest.
    
    .. versionadded:: 1.5.10
    """

    return self._correl_box

  def finish(self) -> None:
    """Method called when the user tries to close the configuration window.

    Checks that a patch was selected on the image. If not, warns the user and
    prevents him from exiting except with CTRL+C.
    
    .. versionadded:: 2.0.0
    """

    if self.box.no_points():
      self.log(logging.WARNING, "No ROI selected ! Not exiting the "
                                "configuration window")
      showerror('Error !',
                message="Please select a ROI before exiting the config "
                        "window !\nOr hit CTRL+C to exit Crappy")
      return

    super().stop()

  def _set_bindings(self) -> None:
    """Binds the left mouse button click for drawing the box on which the
    correlation will be performed."""

    super()._set_bindings()

    self._img_canvas.bind('<ButtonPress-1>', self._start_box)
    self._img_canvas.bind('<B1-Motion>', self._extend_box)
    self._img_canvas.bind('<ButtonRelease-1>', self._stop_box)

  def _start_box(self, event: tk.Event) -> None:
    """Simply saves the position of the user click, and disables the display of
    the current correl box."""

    self.log(logging.DEBUG, "Starting the selection box")

    # If the mouse is on the canvas but not on the image, do nothing
    if not self._check_event_pos(event):
      return

    self._select_box.x_start, \
        self._select_box.y_start = self._coord_to_pix(event.x, event.y)

    self._draw_correl_box = False

  def _stop_box(self, _: tk.Event) -> None:
    """Makes sure that the selected region is valid, sets it as the new correl
    box, and enables the display of the correl box."""

    self.log(logging.DEBUG, "Ending the selection box")

    # If it's just a regular click with no dragging, do nothing
    if self._img is None or self._select_box.no_points():
      self._select_box.reset()
      self._draw_correl_box = True
      return

    # The sides need to be sorted before slicing numpy array
    y_left, y_right, x_top, x_bottom = self._select_box.sorted()

    # If the box is flat, resetting it
    if y_left == y_right or x_top == x_bottom:
      self._select_box.reset()
      self._draw_correl_box = True
      return

    # The new correl box is just the copy of the select box
    self._correl_box.update(self._select_box)
    self._select_box.reset()

    self._draw_correl_box = True

  def _on_img_resize(self, _: Optional[tk.Event] = None) -> None:
    """Same as in the parent class except it also draws the select box on top
    of the displayed image."""

    self.log(logging.DEBUG, "The image canvas was resized")

    # Do not draw the correl box if the user is creating the select box
    if self._draw_correl_box:
      self._draw_box(self._correl_box)
    self._draw_box(self._select_box)

    self._resize_img()
    self._display_img()
    self.update()

  def _update_img(self) -> None:
    """Same as in the parent class except it also draws the select box on top
    of the displayed image."""

    self.log(logging.DEBUG, "Updating the image")

    ret = self._camera.get_image()

    # If no frame could be grabbed from the camera
    if ret is None:
      # If it's the first call, generate error image to initialize the window
      if not self._n_loops:
        self.log(logging.WARNING, "Could not get an image from the camera, "
                                  "displaying an error image instead")
        ret = None, np.array(Image.open(BytesIO(resource_string(
          'crappy', 'tool/data/no_image.png'))))
      # Otherwise, just pass
      else:
        self.log(logging.DEBUG, "No image returned by the camera")
        self.update()
        sleep(0.001)
        return

    self._n_loops += 1
    _, img = ret

    if img.dtype != self.dtype:
      self.dtype = img.dtype
    if self.shape != img.shape:
      self.shape = img.shape

    self._cast_img(img)
    # Do not draw the correl box if the user is creating the select box
    if self._draw_correl_box:
      self._draw_box(self._correl_box)
    self._draw_box(self._select_box)
    self._resize_img()

    self._calc_hist()
    self._resize_hist()

    self._display_img()
    self._display_hist()

    self._update_pixel_value()

    self.update()

  def _handle_box_outside_img(self, _: Box) -> None:
    """If the correl box is outside the image, it means that the image size has
    been modified. Simply resetting the correl box then."""

    self._correl_box.reset()



# coding: utf-8

from typing import Dict, Type

from .meta_modifier import Modifier, MetaModifier

from .demux import Demux
from .differentiate import Diff
from .integrate import Integrate
from .mean import Mean
from .median import Median
from .moving_avg import MovingAvg
from .moving_med import MovingMed
from .offset import Offset
from .trig_on_change import TrigOnChange
from .trig_on_value import TrigOnValue

from ._deprecated import Moving_avg, Moving_med, Trig_on_change, Trig_on_value

modifier_dict: Dict[str, Type[Modifier]] = MetaModifier.classes



# coding: utf-8

from struct import pack_into
from time import sleep
from typing import Union, Optional, Tuple
import logging
from warnings import warn

from ..meta_actuator import Actuator
from ...tool.ft232h import FT232HServer as FT232H, USBArgsType

motor_hat_ctrl = {1: 0x26,
                  2: 0x3A,
                  3: 0x0E,
                  4: 0x22}

motor_hat_pos = {1: 0x2A,
                 2: 0x32,
                 3: 0x12,
                 4: 0x1A}

motor_hat_neg = {1: 0x2E,
                 2: 0x36,
                 3: 0x16,
                 4: 0x1E}

motor_hat_0xFF = [0x00, 0x10, 0x00, 0x00]
motor_hat_max_volt = 12


class DCMotorHatFT232H(Actuator):
  """Class for driving Adafruit's DC motor HAT via an FT232H USB to I2C
  converter.

  It can drive up to four DC motors in speed only. The acquisition of the speed
  has not been implemented so far. It implements the same functionality as the
  :class:`~crappy.actuator.DCMotorHat`, but communicates with the hat over USB
  via an FT232H device.

  Important:
    As this Actuator can drive up to 4 motors simultaneously, it takes a
    :obj:`tuple` as a command, see :meth:`set_speed`. Regular Actuators receive
    their commands as :obj:`float`. A :class:`~crappy.modifier.Modifier` can be
    used for converting a :obj:`float` command from a
    :class:`~crappy.blocks.Generator` to a :obj:`tuple`.

  Note:
    The DC Motor Hat can also drive stepper motors, but this feature isn't
    included here.

  .. versionadded:: 2.0.0
  """

  ft232h = True

  def __init__(self,
               device_address: int = 0x60,
               _ft232h_args: USBArgsType = tuple()) -> None:
    """Checks the validity of the arguments and opens the connection to the
    FT232H.

    Args:
      device_address: The I2C address of the HAT. The default address is
        `0x60`, but it is possible to change this setting by cutting traces on
        the board.
      _ft232h_args: This argument is meant for internal use only and should not
        be provided by the user. It contains the information necessary for
        setting up the FT232H.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._bus: Optional[FT232H] = None
    self._buf: bytearray = bytearray(4)

    super().__init__()

    if not isinstance(device_address, int):
      raise TypeError("device_address should be an integer between 0 and 127.")
    self._address = device_address

    (block_index, block_lock, command_file, answer_file, shared_lock,
     current_block) = _ft232h_args

    self._bus = FT232H(mode='I2C',
                       block_index=block_index,
                       current_block=current_block,
                       command_file=command_file,
                       answer_file=answer_file,
                       block_lock=block_lock,
                       shared_lock=shared_lock)

  def open(self) -> None:
    """Opens the connection to the motor hat and initializes it."""

    self.log(logging.INFO, "Opening with a USB connection over FT232H")

    # Reset
    self._write_i2c(0x00, [0x00])
    sleep(0.01)

    # Sleep & restart, also setting the frequency to 1526Hz
    self._write_i2c(0x00, [0x10])
    self._write_i2c(0xFE, [int(25E6 / 4096.0 / 1600.0 + 0.5)])
    sleep(0.01)
    self._write_i2c(0x00, [0x00])
    sleep(0.01)
    self._write_i2c(0x00, [0xA0])

    # Initializing the motors
    for i in range(1, 5):
      self._write_i2c(motor_hat_ctrl[i], motor_hat_0xFF)

  def set_speed(self, cmd: Tuple[int, float]) -> None:
    """Sets the desired voltage on the selected motor.

    The provided voltage should be between `-12` and `12V` which are the limits
    of the motor hat. If not, it will be silently clamped in this interval.

    Warning:
      Unlike most Actuators, this one takes :obj:`tuple` as commands instead
      of :obj:`float`.

    Args:
      cmd: A :obj:`tuple` containing first the index of the motor to drive as
        an :obj:`int` between 1 and 5, and second the voltage to apply to the
        selected motor as a :obj:`float`."""

    try:
      motor_nr, volts = cmd
    except (TypeError, ValueError):
      raise ValueError("The DCMotorHat receives commands as tuples, with first"
                       " the index of the motor and second the voltage "
                       "command")

    if motor_nr not in range(1, 5):
      raise ValueError("The index of the motor to drive should be an integer "
                       "between 1 and 5")

    volt_clamped = min(abs(volts) / motor_hat_max_volt, 1.0)
    self.log(logging.DEBUG, f"Setting motor {motor_nr} to {volt_clamped}V")
    self._set_motor(motor_nr, volt_clamped)

  def stop(self) -> None:
    """Simply sets the command to `0` to stop the motors."""

    if self._bus is not None:
      for i in range(1, 5):
        self.set_speed((i, 0))

  def close(self) -> None:
    """Closes the I2C connection to the motor hat."""

    if self._bus is not None:
      self.log(logging.INFO, "Closing the connection to the Motor Hat")
      self._bus.close()

  def _set_motor(self, nr: int, cmd: float) -> None:
    """Sets the PWMs associated with a given motor.

    Args:
      nr: The index of the motor to drive.
      cmd: The command as a :obj:`float`, that should be between -1 and 1.
    """

    # Special settings for the extreme values
    if cmd == 0:
      self._write_i2c(motor_hat_pos[nr], motor_hat_0xFF)
      self._write_i2c(motor_hat_neg[nr], motor_hat_0xFF)
    elif cmd == 1.0:
      self._write_i2c(motor_hat_pos[nr], motor_hat_0xFF)
      self._write_i2c(motor_hat_neg[nr], [0x00 for _ in range(4)])
    elif cmd == -1.0:
      self._write_i2c(motor_hat_pos[nr], [0x00 for _ in range(4)])
      self._write_i2c(motor_hat_neg[nr], motor_hat_0xFF)

    # The positive line is driven for positive commands, and reversely
    elif cmd > 0:
      duty_cycle = (int(0xFFFF * cmd) + 1) >> 4
      pack_into("<HH", self._buf, 0, *(0, duty_cycle))
      self._write_i2c(motor_hat_pos[nr], self._buf)
      self._write_i2c(motor_hat_neg[nr], [0x00 for _ in range(4)])
    else:
      duty_cycle = (int(0xFFFF * abs(cmd)) + 1) >> 4
      pack_into("<HH", self._buf, 0, *(0, duty_cycle))
      self._write_i2c(motor_hat_pos[nr], [0x00 for _ in range(4)])
      self._write_i2c(motor_hat_neg[nr], self._buf)

  def _write_i2c(self, register: int, buf: Union[list, bytearray]) -> None:
    """Thin wrapper to reduce verbosity."""

    self._bus.write_i2c_block_data(self._address, register, list(buf))
